from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="VolumeTemplateState")


@_attrs_define
class VolumeTemplateState:
    """Volume template state

    Attributes:
        last_version_uploaded_at (Union[Unset, str]): Timestamp of last version upload
        latest_version (Union[Unset, str]): Current/latest S3 version ID
        status (Union[Unset, str]): Status of the volume template (created, ready, error)
        version_count (Union[Unset, int]): Total number of versions for this template
    """

    last_version_uploaded_at: Union[Unset, str] = UNSET
    latest_version: Union[Unset, str] = UNSET
    status: Union[Unset, str] = UNSET
    version_count: Union[Unset, int] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        last_version_uploaded_at = self.last_version_uploaded_at

        latest_version = self.latest_version

        status = self.status

        version_count = self.version_count

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if last_version_uploaded_at is not UNSET:
            field_dict["lastVersionUploadedAt"] = last_version_uploaded_at
        if latest_version is not UNSET:
            field_dict["latestVersion"] = latest_version
        if status is not UNSET:
            field_dict["status"] = status
        if version_count is not UNSET:
            field_dict["versionCount"] = version_count

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        last_version_uploaded_at = d.pop("lastVersionUploadedAt", UNSET)

        latest_version = d.pop("latestVersion", UNSET)

        status = d.pop("status", UNSET)

        version_count = d.pop("versionCount", UNSET)

        volume_template_state = cls(
            last_version_uploaded_at=last_version_uploaded_at,
            latest_version=latest_version,
            status=status,
            version_count=version_count,
        )

        volume_template_state.additional_properties = d
        return volume_template_state

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.core_spec_configurations import CoreSpecConfigurations
    from ..models.flavor import Flavor
    from ..models.model_private_cluster import ModelPrivateCluster
    from ..models.revision_configuration import RevisionConfiguration
    from ..models.runtime import Runtime
    from ..models.trigger import Trigger


T = TypeVar("T", bound="JobSpec")


@_attrs_define
class JobSpec:
    """Job specification

    Attributes:
        configurations (Union[Unset, CoreSpecConfigurations]): Optional configurations for the object
        enabled (Union[Unset, bool]): Enable or disable the resource
        flavors (Union[Unset, list['Flavor']]): Types of hardware available for deployments
        integration_connections (Union[Unset, list[str]]):
        policies (Union[Unset, list[str]]):
        private_clusters (Union[Unset, ModelPrivateCluster]): Private cluster where the model deployment is deployed
        revision (Union[Unset, RevisionConfiguration]): Revision configuration
        runtime (Union[Unset, Runtime]): Set of configurations for a deployment
        sandbox (Union[Unset, bool]): Sandbox mode
        region (Union[Unset, str]): Region where the job should be created (e.g. us-was-1, eu-lon-1)
        triggers (Union[Unset, list['Trigger']]): Triggers to use your agent
    """

    configurations: Union[Unset, "CoreSpecConfigurations"] = UNSET
    enabled: Union[Unset, bool] = UNSET
    flavors: Union[Unset, list["Flavor"]] = UNSET
    integration_connections: Union[Unset, list[str]] = UNSET
    policies: Union[Unset, list[str]] = UNSET
    private_clusters: Union[Unset, "ModelPrivateCluster"] = UNSET
    revision: Union[Unset, "RevisionConfiguration"] = UNSET
    runtime: Union[Unset, "Runtime"] = UNSET
    sandbox: Union[Unset, bool] = UNSET
    region: Union[Unset, str] = UNSET
    triggers: Union[Unset, list["Trigger"]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        configurations: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.configurations
            and not isinstance(self.configurations, Unset)
            and not isinstance(self.configurations, dict)
        ):
            configurations = self.configurations.to_dict()
        elif self.configurations and isinstance(self.configurations, dict):
            configurations = self.configurations

        enabled = self.enabled

        flavors: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.flavors, Unset):
            flavors = []
            for componentsschemas_flavors_item_data in self.flavors:
                if type(componentsschemas_flavors_item_data) is dict:
                    componentsschemas_flavors_item = componentsschemas_flavors_item_data
                else:
                    componentsschemas_flavors_item = componentsschemas_flavors_item_data.to_dict()
                flavors.append(componentsschemas_flavors_item)

        integration_connections: Union[Unset, list[str]] = UNSET
        if not isinstance(self.integration_connections, Unset):
            integration_connections = self.integration_connections

        policies: Union[Unset, list[str]] = UNSET
        if not isinstance(self.policies, Unset):
            policies = self.policies

        private_clusters: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.private_clusters
            and not isinstance(self.private_clusters, Unset)
            and not isinstance(self.private_clusters, dict)
        ):
            private_clusters = self.private_clusters.to_dict()
        elif self.private_clusters and isinstance(self.private_clusters, dict):
            private_clusters = self.private_clusters

        revision: Union[Unset, dict[str, Any]] = UNSET
        if self.revision and not isinstance(self.revision, Unset) and not isinstance(self.revision, dict):
            revision = self.revision.to_dict()
        elif self.revision and isinstance(self.revision, dict):
            revision = self.revision

        runtime: Union[Unset, dict[str, Any]] = UNSET
        if self.runtime and not isinstance(self.runtime, Unset) and not isinstance(self.runtime, dict):
            runtime = self.runtime.to_dict()
        elif self.runtime and isinstance(self.runtime, dict):
            runtime = self.runtime

        sandbox = self.sandbox

        region = self.region

        triggers: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.triggers, Unset):
            triggers = []
            for componentsschemas_triggers_item_data in self.triggers:
                if type(componentsschemas_triggers_item_data) is dict:
                    componentsschemas_triggers_item = componentsschemas_triggers_item_data
                else:
                    componentsschemas_triggers_item = componentsschemas_triggers_item_data.to_dict()
                triggers.append(componentsschemas_triggers_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if configurations is not UNSET:
            field_dict["configurations"] = configurations
        if enabled is not UNSET:
            field_dict["enabled"] = enabled
        if flavors is not UNSET:
            field_dict["flavors"] = flavors
        if integration_connections is not UNSET:
            field_dict["integrationConnections"] = integration_connections
        if policies is not UNSET:
            field_dict["policies"] = policies
        if private_clusters is not UNSET:
            field_dict["privateClusters"] = private_clusters
        if revision is not UNSET:
            field_dict["revision"] = revision
        if runtime is not UNSET:
            field_dict["runtime"] = runtime
        if sandbox is not UNSET:
            field_dict["sandbox"] = sandbox
        if region is not UNSET:
            field_dict["region"] = region
        if triggers is not UNSET:
            field_dict["triggers"] = triggers

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.core_spec_configurations import CoreSpecConfigurations
        from ..models.flavor import Flavor
        from ..models.model_private_cluster import ModelPrivateCluster
        from ..models.revision_configuration import RevisionConfiguration
        from ..models.runtime import Runtime
        from ..models.trigger import Trigger

        if not src_dict:
            return None
        d = src_dict.copy()
        _configurations = d.pop("configurations", UNSET)
        configurations: Union[Unset, CoreSpecConfigurations]
        if isinstance(_configurations, Unset):
            configurations = UNSET
        else:
            configurations = CoreSpecConfigurations.from_dict(_configurations)

        enabled = d.pop("enabled", UNSET)

        flavors = []
        _flavors = d.pop("flavors", UNSET)
        for componentsschemas_flavors_item_data in _flavors or []:
            componentsschemas_flavors_item = Flavor.from_dict(componentsschemas_flavors_item_data)

            flavors.append(componentsschemas_flavors_item)

        integration_connections = cast(list[str], d.pop("integrationConnections", UNSET))

        policies = cast(list[str], d.pop("policies", UNSET))

        _private_clusters = d.pop("privateClusters", UNSET)
        private_clusters: Union[Unset, ModelPrivateCluster]
        if isinstance(_private_clusters, Unset):
            private_clusters = UNSET
        else:
            private_clusters = ModelPrivateCluster.from_dict(_private_clusters)

        _revision = d.pop("revision", UNSET)
        revision: Union[Unset, RevisionConfiguration]
        if isinstance(_revision, Unset):
            revision = UNSET
        else:
            revision = RevisionConfiguration.from_dict(_revision)

        _runtime = d.pop("runtime", UNSET)
        runtime: Union[Unset, Runtime]
        if isinstance(_runtime, Unset):
            runtime = UNSET
        else:
            runtime = Runtime.from_dict(_runtime)

        sandbox = d.pop("sandbox", UNSET)

        region = d.pop("region", UNSET)

        triggers = []
        _triggers = d.pop("triggers", UNSET)
        for componentsschemas_triggers_item_data in _triggers or []:
            componentsschemas_triggers_item = Trigger.from_dict(componentsschemas_triggers_item_data)

            triggers.append(componentsschemas_triggers_item)

        job_spec = cls(
            configurations=configurations,
            enabled=enabled,
            flavors=flavors,
            integration_connections=integration_connections,
            policies=policies,
            private_clusters=private_clusters,
            revision=revision,
            runtime=runtime,
            sandbox=sandbox,
            region=region,
            triggers=triggers,
        )

        job_spec.additional_properties = d
        return job_spec

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="JobExecutionTaskCondition")


@_attrs_define
class JobExecutionTaskCondition:
    """Job execution task condition

    Attributes:
        execution_reason (Union[Unset, str]): Execution reason
        message (Union[Unset, str]): Condition message
        reason (Union[Unset, str]): Condition reason
        severity (Union[Unset, str]): Condition severity
        state (Union[Unset, str]): Condition state
        type_ (Union[Unset, str]): Condition type
    """

    execution_reason: Union[Unset, str] = UNSET
    message: Union[Unset, str] = UNSET
    reason: Union[Unset, str] = UNSET
    severity: Union[Unset, str] = UNSET
    state: Union[Unset, str] = UNSET
    type_: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        execution_reason = self.execution_reason

        message = self.message

        reason = self.reason

        severity = self.severity

        state = self.state

        type_ = self.type_

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if execution_reason is not UNSET:
            field_dict["executionReason"] = execution_reason
        if message is not UNSET:
            field_dict["message"] = message
        if reason is not UNSET:
            field_dict["reason"] = reason
        if severity is not UNSET:
            field_dict["severity"] = severity
        if state is not UNSET:
            field_dict["state"] = state
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        execution_reason = d.pop("executionReason", UNSET)

        message = d.pop("message", UNSET)

        reason = d.pop("reason", UNSET)

        severity = d.pop("severity", UNSET)

        state = d.pop("state", UNSET)

        type_ = d.pop("type", UNSET)

        job_execution_task_condition = cls(
            execution_reason=execution_reason,
            message=message,
            reason=reason,
            severity=severity,
            state=state,
            type_=type_,
        )

        job_execution_task_condition.additional_properties = d
        return job_execution_task_condition

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import logging

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.trace import Tracer, get_tracer
from wrapt import wrap_function_wrapper

from blaxel.telemetry.span import SpanManager

logger = logging.getLogger(__name__)


class BlaxelLanggraphInstrumentor(BaseInstrumentor):
    def instrumentation_dependencies(self):
        return ["langgraph"]

    def _instrument(self, **kwargs):
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, "0.2.0", tracer_provider)
        # self._patch_gemini(tracer)
        self._patch_tool(tracer)

    def _uninstrument(self, **kwargs):
        import blaxel.langgraph.custom.gemini

        # Restore original methods
        if hasattr(
            blaxel.langgraph.custom.gemini.GeminiRestClient, "_blaxel_original_generate_content"
        ):
            blaxel.langgraph.custom.gemini.GeminiRestClient.generate_content = (
                blaxel.langgraph.custom.gemini.GeminiRestClient._blaxel_original_generate_content
            )
            delattr(
                blaxel.langgraph.custom.gemini.GeminiRestClient, "_blaxel_original_generate_content"
            )

        if hasattr(
            blaxel.langgraph.custom.gemini.GeminiRestClient,
            "_blaxel_original_generate_content_async",
        ):
            blaxel.langgraph.custom.gemini.GeminiRestClient.generate_content_async = blaxel.langgraph.custom.gemini.GeminiRestClient._blaxel_original_generate_content_async
            delattr(
                blaxel.langgraph.custom.gemini.GeminiRestClient,
                "_blaxel_original_generate_content_async",
            )

        if hasattr(
            blaxel.langgraph.custom.gemini.GeminiRestClient,
            "_blaxel_original_stream_generate_content",
        ):
            blaxel.langgraph.custom.gemini.GeminiRestClient.stream_generate_content = blaxel.langgraph.custom.gemini.GeminiRestClient._blaxel_original_stream_generate_content
            delattr(
                blaxel.langgraph.custom.gemini.GeminiRestClient,
                "_blaxel_original_stream_generate_content",
            )

        if hasattr(
            blaxel.langgraph.custom.gemini.GeminiRestClient,
            "_blaxel_original_stream_generate_content_async",
        ):
            blaxel.langgraph.custom.gemini.GeminiRestClient.stream_generate_content_async = blaxel.langgraph.custom.gemini.GeminiRestClient._blaxel_original_stream_generate_content_async
            delattr(
                blaxel.langgraph.custom.gemini.GeminiRestClient,
                "_blaxel_original_stream_generate_content_async",
            )

    def _patch_tool(self, tracer: Tracer):
        def _traced_run(original, instance, args, kwargs):
            attributes = {
                "tool.type": "langgraph",
                "span.type": "tool.call",
                **SpanManager.get_default_attributes(),
            }
            with tracer.start_span("langgraph.tool", attributes=attributes) as span:
                try:
                    result = original(*args, **kwargs)
                    span.set_attribute("tool.name", result.name)
                    return result
                except Exception as e:
                    span.set_attribute("tool.error", str(e))
                    raise

        async def _traced_arun(original, instance, args, kwargs):
            attributes = {
                "tool.type": "langgraph",
                "span.type": "tool.call",
                **SpanManager.get_default_attributes(),
            }
            with tracer.start_span("langgraph.tool", attributes=attributes) as span:
                try:
                    result = await original(*args, **kwargs)
                    span.set_attribute("tool.name", result.name)
                    return result
                except Exception as e:
                    span.set_attribute("tool.error", str(e))
                    raise

        wrap_function_wrapper("langchain_core.tools", "BaseTool.run", _traced_run)
        wrap_function_wrapper("langchain_core.tools", "BaseTool.arun", _traced_arun)
        wrap_function_wrapper("langchain_core.tools", "BaseTool.arun", _traced_arun)



from ..types import SandboxConfiguration
from .action import SyncSandboxAction


class SyncSandboxNetwork(SyncSandboxAction):
    def __init__(self, sandbox_config: SandboxConfiguration):
        super().__init__(sandbox_config)

    # Placeholder for future sync network operations





from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.job_execution_metadata import JobExecutionMetadata
    from ..models.job_execution_spec import JobExecutionSpec
    from ..models.job_execution_stats import JobExecutionStats
    from ..models.job_execution_task import JobExecutionTask


T = TypeVar("T", bound="JobExecution")


@_attrs_define
class JobExecution:
    """Job execution

    Attributes:
        metadata (Union[Unset, JobExecutionMetadata]): Job execution metadata
        spec (Union[Unset, JobExecutionSpec]): Job execution specification
        stats (Union[Unset, JobExecutionStats]): Job execution statistics
        status (Union[Unset, str]): Job execution status
        tasks (Union[Unset, list['JobExecutionTask']]): List of execution tasks
    """

    metadata: Union[Unset, "JobExecutionMetadata"] = UNSET
    spec: Union[Unset, "JobExecutionSpec"] = UNSET
    stats: Union[Unset, "JobExecutionStats"] = UNSET
    status: Union[Unset, str] = UNSET
    tasks: Union[Unset, list["JobExecutionTask"]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        metadata: Union[Unset, dict[str, Any]] = UNSET
        if self.metadata and not isinstance(self.metadata, Unset) and not isinstance(self.metadata, dict):
            metadata = self.metadata.to_dict()
        elif self.metadata and isinstance(self.metadata, dict):
            metadata = self.metadata

        spec: Union[Unset, dict[str, Any]] = UNSET
        if self.spec and not isinstance(self.spec, Unset) and not isinstance(self.spec, dict):
            spec = self.spec.to_dict()
        elif self.spec and isinstance(self.spec, dict):
            spec = self.spec

        stats: Union[Unset, dict[str, Any]] = UNSET
        if self.stats and not isinstance(self.stats, Unset) and not isinstance(self.stats, dict):
            stats = self.stats.to_dict()
        elif self.stats and isinstance(self.stats, dict):
            stats = self.stats

        status = self.status

        tasks: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.tasks, Unset):
            tasks = []
            for tasks_item_data in self.tasks:
                if type(tasks_item_data) is dict:
                    tasks_item = tasks_item_data
                else:
                    tasks_item = tasks_item_data.to_dict()
                tasks.append(tasks_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if metadata is not UNSET:
            field_dict["metadata"] = metadata
        if spec is not UNSET:
            field_dict["spec"] = spec
        if stats is not UNSET:
            field_dict["stats"] = stats
        if status is not UNSET:
            field_dict["status"] = status
        if tasks is not UNSET:
            field_dict["tasks"] = tasks

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.job_execution_metadata import JobExecutionMetadata
        from ..models.job_execution_spec import JobExecutionSpec
        from ..models.job_execution_stats import JobExecutionStats
        from ..models.job_execution_task import JobExecutionTask

        if not src_dict:
            return None
        d = src_dict.copy()
        _metadata = d.pop("metadata", UNSET)
        metadata: Union[Unset, JobExecutionMetadata]
        if isinstance(_metadata, Unset):
            metadata = UNSET
        else:
            metadata = JobExecutionMetadata.from_dict(_metadata)

        _spec = d.pop("spec", UNSET)
        spec: Union[Unset, JobExecutionSpec]
        if isinstance(_spec, Unset):
            spec = UNSET
        else:
            spec = JobExecutionSpec.from_dict(_spec)

        _stats = d.pop("stats", UNSET)
        stats: Union[Unset, JobExecutionStats]
        if isinstance(_stats, Unset):
            stats = UNSET
        else:
            stats = JobExecutionStats.from_dict(_stats)

        status = d.pop("status", UNSET)

        tasks = []
        _tasks = d.pop("tasks", UNSET)
        for tasks_item_data in _tasks or []:
            tasks_item = JobExecutionTask.from_dict(tasks_item_data)

            tasks.append(tasks_item)

        job_execution = cls(
            metadata=metadata,
            spec=spec,
            stats=stats,
            status=status,
            tasks=tasks,
        )

        job_execution.additional_properties = d
        return job_execution

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.continent import Continent
    from ..models.country import Country
    from ..models.private_location import PrivateLocation
    from ..models.region import Region


T = TypeVar("T", bound="Configuration")


@_attrs_define
class Configuration:
    """Configuration

    Attributes:
        continents (Union[Unset, list['Continent']]): Continents
        countries (Union[Unset, list['Country']]): Countries
        private_locations (Union[Unset, list['PrivateLocation']]): Private locations managed with blaxel operator
        regions (Union[Unset, list['Region']]): Regions
    """

    continents: Union[Unset, list["Continent"]] = UNSET
    countries: Union[Unset, list["Country"]] = UNSET
    private_locations: Union[Unset, list["PrivateLocation"]] = UNSET
    regions: Union[Unset, list["Region"]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        continents: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.continents, Unset):
            continents = []
            for continents_item_data in self.continents:
                if type(continents_item_data) is dict:
                    continents_item = continents_item_data
                else:
                    continents_item = continents_item_data.to_dict()
                continents.append(continents_item)

        countries: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.countries, Unset):
            countries = []
            for countries_item_data in self.countries:
                if type(countries_item_data) is dict:
                    countries_item = countries_item_data
                else:
                    countries_item = countries_item_data.to_dict()
                countries.append(countries_item)

        private_locations: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.private_locations, Unset):
            private_locations = []
            for private_locations_item_data in self.private_locations:
                if type(private_locations_item_data) is dict:
                    private_locations_item = private_locations_item_data
                else:
                    private_locations_item = private_locations_item_data.to_dict()
                private_locations.append(private_locations_item)

        regions: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.regions, Unset):
            regions = []
            for regions_item_data in self.regions:
                if type(regions_item_data) is dict:
                    regions_item = regions_item_data
                else:
                    regions_item = regions_item_data.to_dict()
                regions.append(regions_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if continents is not UNSET:
            field_dict["continents"] = continents
        if countries is not UNSET:
            field_dict["countries"] = countries
        if private_locations is not UNSET:
            field_dict["privateLocations"] = private_locations
        if regions is not UNSET:
            field_dict["regions"] = regions

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.continent import Continent
        from ..models.country import Country
        from ..models.private_location import PrivateLocation
        from ..models.region import Region

        if not src_dict:
            return None
        d = src_dict.copy()
        continents = []
        _continents = d.pop("continents", UNSET)
        for continents_item_data in _continents or []:
            continents_item = Continent.from_dict(continents_item_data)

            continents.append(continents_item)

        countries = []
        _countries = d.pop("countries", UNSET)
        for countries_item_data in _countries or []:
            countries_item = Country.from_dict(countries_item_data)

            countries.append(countries_item)

        private_locations = []
        _private_locations = d.pop("privateLocations", UNSET)
        for private_locations_item_data in _private_locations or []:
            private_locations_item = PrivateLocation.from_dict(private_locations_item_data)

            private_locations.append(private_locations_item)

        regions = []
        _regions = d.pop("regions", UNSET)
        for regions_item_data in _regions or []:
            regions_item = Region.from_dict(regions_item_data)

            regions.append(regions_item)

        configuration = cls(
            continents=continents,
            countries=countries,
            private_locations=private_locations,
            regions=regions,
        )

        configuration.additional_properties = d
        return configuration

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.create_job_execution_request_tasks_item import CreateJobExecutionRequestTasksItem


T = TypeVar("T", bound="CreateJobExecutionRequest")


@_attrs_define
class CreateJobExecutionRequest:
    """Request to create a job execution

    Attributes:
        execution_id (Union[Unset, str]): Execution ID (optional, will be generated if not provided)
        id (Union[Unset, str]): Unique message ID
        job_id (Union[Unset, str]): Job ID
        tasks (Union[Unset, list['CreateJobExecutionRequestTasksItem']]): Array of task parameters for parallel
            execution
        workspace_id (Union[Unset, str]): Workspace ID
    """

    execution_id: Union[Unset, str] = UNSET
    id: Union[Unset, str] = UNSET
    job_id: Union[Unset, str] = UNSET
    tasks: Union[Unset, list["CreateJobExecutionRequestTasksItem"]] = UNSET
    workspace_id: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        execution_id = self.execution_id

        id = self.id

        job_id = self.job_id

        tasks: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.tasks, Unset):
            tasks = []
            for tasks_item_data in self.tasks:
                if type(tasks_item_data) is dict:
                    tasks_item = tasks_item_data
                else:
                    tasks_item = tasks_item_data.to_dict()
                tasks.append(tasks_item)

        workspace_id = self.workspace_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if execution_id is not UNSET:
            field_dict["executionId"] = execution_id
        if id is not UNSET:
            field_dict["id"] = id
        if job_id is not UNSET:
            field_dict["jobId"] = job_id
        if tasks is not UNSET:
            field_dict["tasks"] = tasks
        if workspace_id is not UNSET:
            field_dict["workspaceId"] = workspace_id

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.create_job_execution_request_tasks_item import (
            CreateJobExecutionRequestTasksItem,
        )

        if not src_dict:
            return None
        d = src_dict.copy()
        execution_id = d.pop("executionId", UNSET)

        id = d.pop("id", UNSET)

        job_id = d.pop("jobId", UNSET)

        tasks = []
        _tasks = d.pop("tasks", UNSET)
        for tasks_item_data in _tasks or []:
            tasks_item = CreateJobExecutionRequestTasksItem.from_dict(tasks_item_data)

            tasks.append(tasks_item)

        workspace_id = d.pop("workspaceId", UNSET)

        create_job_execution_request = cls(
            execution_id=execution_id,
            id=id,
            job_id=job_id,
            tasks=tasks,
            workspace_id=workspace_id,
        )

        create_job_execution_request.additional_properties = d
        return create_job_execution_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="JobExecutionMetadata")


@_attrs_define
class JobExecutionMetadata:
    """Job execution metadata

    Attributes:
        cluster (Union[Unset, str]): Cluster ID
        completed_at (Union[Unset, str]): Completion timestamp
        created_at (Union[Unset, str]): Creation timestamp
        deleted_at (Union[Unset, str]): Deletion timestamp
        expired_at (Union[Unset, str]): Expiration timestamp
        id (Union[Unset, str]): Execution ID
        job (Union[Unset, str]): Job name
        started_at (Union[Unset, str]): Start timestamp
        updated_at (Union[Unset, str]): Last update timestamp
        workspace (Union[Unset, str]): Workspace ID
    """

    cluster: Union[Unset, str] = UNSET
    completed_at: Union[Unset, str] = UNSET
    created_at: Union[Unset, str] = UNSET
    deleted_at: Union[Unset, str] = UNSET
    expired_at: Union[Unset, str] = UNSET
    id: Union[Unset, str] = UNSET
    job: Union[Unset, str] = UNSET
    started_at: Union[Unset, str] = UNSET
    updated_at: Union[Unset, str] = UNSET
    workspace: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        cluster = self.cluster

        completed_at = self.completed_at

        created_at = self.created_at

        deleted_at = self.deleted_at

        expired_at = self.expired_at

        id = self.id

        job = self.job

        started_at = self.started_at

        updated_at = self.updated_at

        workspace = self.workspace

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if cluster is not UNSET:
            field_dict["cluster"] = cluster
        if completed_at is not UNSET:
            field_dict["completedAt"] = completed_at
        if created_at is not UNSET:
            field_dict["createdAt"] = created_at
        if deleted_at is not UNSET:
            field_dict["deletedAt"] = deleted_at
        if expired_at is not UNSET:
            field_dict["expiredAt"] = expired_at
        if id is not UNSET:
            field_dict["id"] = id
        if job is not UNSET:
            field_dict["job"] = job
        if started_at is not UNSET:
            field_dict["startedAt"] = started_at
        if updated_at is not UNSET:
            field_dict["updatedAt"] = updated_at
        if workspace is not UNSET:
            field_dict["workspace"] = workspace

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        cluster = d.pop("cluster", UNSET)

        completed_at = d.pop("completedAt", UNSET)

        created_at = d.pop("createdAt", UNSET)

        deleted_at = d.pop("deletedAt", UNSET)

        expired_at = d.pop("expiredAt", UNSET)

        id = d.pop("id", UNSET)

        job = d.pop("job", UNSET)

        started_at = d.pop("startedAt", UNSET)

        updated_at = d.pop("updatedAt", UNSET)

        workspace = d.pop("workspace", UNSET)

        job_execution_metadata = cls(
            cluster=cluster,
            completed_at=completed_at,
            created_at=created_at,
            deleted_at=deleted_at,
            expired_at=expired_at,
            id=id,
            job=job,
            started_at=started_at,
            updated_at=updated_at,
            workspace=workspace,
        )

        job_execution_metadata.additional_properties = d
        return job_execution_metadata

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from .interpreter import SyncCodeInterpreter
from .sandbox import (
    SyncSandboxCodegen,
    SyncSandboxFileSystem,
    SyncSandboxInstance,
    SyncSandboxPreviews,
    SyncSandboxProcess,
)

__all__ = [
    "SyncSandboxInstance",
    "SyncSandboxFileSystem",
    "SyncSandboxPreviews",
    "SyncSandboxProcess",
    "SyncSandboxCodegen",
    "SyncCodeInterpreter",
]





from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.trigger_configuration import TriggerConfiguration


T = TypeVar("T", bound="Trigger")


@_attrs_define
class Trigger:
    """Trigger configuration

    Attributes:
        configuration (Union[Unset, TriggerConfiguration]): Trigger configuration
        id (Union[Unset, str]): The id of the trigger
        type_ (Union[Unset, str]): The type of trigger, can be http or http-async
    """

    configuration: Union[Unset, "TriggerConfiguration"] = UNSET
    id: Union[Unset, str] = UNSET
    type_: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        configuration: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.configuration
            and not isinstance(self.configuration, Unset)
            and not isinstance(self.configuration, dict)
        ):
            configuration = self.configuration.to_dict()
        elif self.configuration and isinstance(self.configuration, dict):
            configuration = self.configuration

        id = self.id

        type_ = self.type_

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if configuration is not UNSET:
            field_dict["configuration"] = configuration
        if id is not UNSET:
            field_dict["id"] = id
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.trigger_configuration import TriggerConfiguration

        if not src_dict:
            return None
        d = src_dict.copy()
        _configuration = d.pop("configuration", UNSET)
        configuration: Union[Unset, TriggerConfiguration]
        if isinstance(_configuration, Unset):
            configuration = UNSET
        else:
            configuration = TriggerConfiguration.from_dict(_configuration)

        id = d.pop("id", UNSET)

        type_ = d.pop("type", UNSET)

        trigger = cls(
            configuration=configuration,
            id=id,
            type_=type_,
        )

        trigger.additional_properties = d
        return trigger

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="File")


@_attrs_define
class File:
    """
    Attributes:
        group (str):
        last_modified (str):
        name (str):
        owner (str):
        path (str):
        permissions (str):
        size (int):
    """

    group: str
    last_modified: str
    name: str
    owner: str
    path: str
    permissions: str
    size: int
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        group = self.group

        last_modified = self.last_modified

        name = self.name

        owner = self.owner

        path = self.path

        permissions = self.permissions

        size = self.size

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "group": group,
                "lastModified": last_modified,
                "name": name,
                "owner": owner,
                "path": path,
                "permissions": permissions,
                "size": size,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        group = d.pop("group")

        last_modified = d.pop("lastModified")

        name = d.pop("name")

        owner = d.pop("owner")

        path = d.pop("path")

        permissions = d.pop("permissions")

        size = d.pop("size")

        file = cls(
            group=group,
            last_modified=last_modified,
            name=name,
            owner=owner,
            path=path,
            permissions=permissions,
            size=size,
        )

        file.additional_properties = d
        return file

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import os

# Transformers is a dependency of DeepSeek, and it logs a lot of warnings that are not useful
os.environ["TRANSFORMERS_NO_ADVISORY_WARNINGS"] = "1"

from logging import getLogger
from typing import Any, Sequence

from google.genai.types import HttpOptions
from llama_index.core.base.llms.types import (
    ChatMessage,
    ChatResponse,
    ChatResponseAsyncGen,
    ChatResponseGen,
    CompletionResponse,
    CompletionResponseAsyncGen,
    CompletionResponseGen,
)
from llama_index.llms.anthropic import Anthropic
from llama_index.llms.cerebras import Cerebras
from llama_index.llms.deepseek import DeepSeek
from llama_index.llms.google_genai import GoogleGenAI
from llama_index.llms.groq import Groq
from llama_index.llms.mistralai import MistralAI
from llama_index.llms.openai import OpenAI

from blaxel.core import bl_model as bl_model_core
from blaxel.core import settings

from .custom.cohere import Cohere

logger = getLogger(__name__)


class TokenRefreshingWrapper:
    """Base wrapper class that refreshes token before each call."""
    
    def __init__(self, model_config: dict):
        self.model_config = model_config
        self.wrapped_model = self._create_model()
    
    def _create_model(self):
        """Create the model instance with current token."""
        config = self.model_config
        model_type = config['type']
        model = config['model']
        url = config['url']
        kwargs = config.get('kwargs', {})
        
        if model_type == 'anthropic':
            return Anthropic(
                model=model,
                api_key=settings.auth.token,
                base_url=url,
                default_headers=settings.auth.get_headers(),
                **kwargs
            )
        elif model_type == 'xai':
            return Groq(
                model=model,
                api_key=settings.auth.token,
                api_base=f"{url}/v1",
                **kwargs
            )
        elif model_type == 'gemini':
            return GoogleGenAI(
                api_key=settings.auth.token,
                model=model,
                api_base=f"{url}/v1",
                http_options=HttpOptions(
                    base_url=url,
                    headers=settings.auth.get_headers(),
                ),
                **kwargs
            )
        elif model_type == 'cohere':
            return Cohere(
                model=model,
                api_key=settings.auth.token,
                api_base=url,
                **kwargs
            )
        elif model_type == 'deepseek':
            return DeepSeek(
                model=model,
                api_key=settings.auth.token,
                api_base=f"{url}/v1",
                **kwargs
            )
        elif model_type == 'mistral':
            return MistralAI(
                model=model,
                api_key=settings.auth.token,
                endpoint=url,
                **kwargs
            )
        elif model_type == 'cerebras':
            return Cerebras(
                model=model,
                api_key=settings.auth.token,
                api_base=f"{url}/v1",
                **kwargs
            )
        else:
            if model_type != "openai":
                logger.warning(f"Model {model} is not supported by LlamaIndex, defaulting to OpenAI")
            
            return OpenAI(
                model=model,
                api_key=settings.auth.token,
                api_base=f"{url}/v1",
                **kwargs
            )
    
    def _refresh_token(self):
        """Refresh the token and recreate the model if needed."""
        # Only refresh if using ClientCredentials (which has get_token method)
        current_token = settings.auth.token
        
        if hasattr(settings.auth, 'get_token'):
            # This will trigger token refresh if needed
            settings.auth.get_token()
        
        new_token = settings.auth.token
        
        # If token changed, recreate the model
        if current_token != new_token:
            self.wrapped_model = self._create_model()
    
    def __getattr__(self, name):
        """Delegate attribute access to wrapped model."""
        return getattr(self.wrapped_model, name)


class TokenRefreshingLLM(TokenRefreshingWrapper):
    """Wrapper for LlamaIndex LLMs that refreshes token before each call."""
    
    async def achat(
        self,
        messages: Sequence[ChatMessage],
        **kwargs: Any,
    ) -> ChatResponse:
        """Async chat with token refresh."""
        self._refresh_token()
        return await self.wrapped_model.achat(messages, **kwargs)
    
    def chat(
        self,
        messages: Sequence[ChatMessage],
        **kwargs: Any,
    ) -> ChatResponse:
        """Sync chat with token refresh."""
        self._refresh_token()
        return self.wrapped_model.chat(messages, **kwargs)
    
    async def astream_chat(
        self,
        messages: Sequence[ChatMessage],
        **kwargs: Any,
    ) -> ChatResponseAsyncGen:
        """Async stream chat with token refresh."""
        self._refresh_token()
        async for chunk in self.wrapped_model.astream_chat(messages, **kwargs):
            yield chunk
    
    def stream_chat(
        self,
        messages: Sequence[ChatMessage],
        **kwargs: Any,
    ) -> ChatResponseGen:
        """Sync stream chat with token refresh."""
        self._refresh_token()
        for chunk in self.wrapped_model.stream_chat(messages, **kwargs):
            yield chunk
    
    async def acomplete(
        self,
        prompt: str,
        **kwargs: Any,
    ) -> CompletionResponse:
        """Async complete with token refresh."""
        self._refresh_token()
        return await self.wrapped_model.acomplete(prompt, **kwargs)
    
    def complete(
        self,
        prompt: str,
        **kwargs: Any,
    ) -> CompletionResponse:
        """Sync complete with token refresh."""
        self._refresh_token()
        return self.wrapped_model.complete(prompt, **kwargs)
    
    async def astream_complete(
        self,
        prompt: str,
        **kwargs: Any,
    ) -> CompletionResponseAsyncGen:
        """Async stream complete with token refresh."""
        self._refresh_token()
        async for chunk in self.wrapped_model.astream_complete(prompt, **kwargs):
            yield chunk
    
    def stream_complete(
        self,
        prompt: str,
        **kwargs: Any,
    ) -> CompletionResponseGen:
        """Sync stream complete with token refresh."""
        self._refresh_token()
        for chunk in self.wrapped_model.stream_complete(prompt, **kwargs):
            yield chunk


async def bl_model(name, **kwargs):
    url, type, model = await bl_model_core(name).get_parameters()
    
    # Store model configuration for recreation
    model_config = {
        'type': type,
        'model': model,
        'url': url,
        'kwargs': kwargs
    }
    
    # Create and return the wrapper
    return TokenRefreshingLLM(model_config)


from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.filesystem_uploaded_part import FilesystemUploadedPart


T = TypeVar("T", bound="MultipartListPartsResponse")


@_attrs_define
class MultipartListPartsResponse:
    """
    Attributes:
        parts (Union[Unset, list['FilesystemUploadedPart']]):
        upload_id (Union[Unset, str]):  Example: 550e8400-e29b-41d4-a716-446655440000.
    """

    parts: Union[Unset, list["FilesystemUploadedPart"]] = UNSET
    upload_id: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        parts: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.parts, Unset):
            parts = []
            for parts_item_data in self.parts:
                if type(parts_item_data) is dict:
                    parts_item = parts_item_data
                else:
                    parts_item = parts_item_data.to_dict()
                parts.append(parts_item)

        upload_id = self.upload_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if parts is not UNSET:
            field_dict["parts"] = parts
        if upload_id is not UNSET:
            field_dict["uploadId"] = upload_id

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.filesystem_uploaded_part import FilesystemUploadedPart

        if not src_dict:
            return None
        d = src_dict.copy()
        parts = []
        _parts = d.pop("parts", UNSET)
        for parts_item_data in _parts or []:
            parts_item = FilesystemUploadedPart.from_dict(parts_item_data)

            parts.append(parts_item)

        upload_id = d.pop("uploadId", UNSET)

        multipart_list_parts_response = cls(
            parts=parts,
            upload_id=upload_id,
        )

        multipart_list_parts_response.additional_properties = d
        return multipart_list_parts_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import hashlib
from logging import getLogger

from .env import env

logger = getLogger(__name__)


def get_alphanumeric_limited_hash(input_str, max_size=48):
    """
    Create an alphanumeric hash using MD5 that can be reproduced in Go, TypeScript, and Python.

    Args:
        input_str (str): The input string to hash
        max_size (int): The maximum length of the returned hash

    Returns:
        str: An alphanumeric hash of the input string, limited to max_size
    """
    # Calculate MD5 hash and convert to hexadecimal
    hash_hex = hashlib.md5(input_str.encode("utf-8")).hexdigest()

    # Limit to max_size
    if len(hash_hex) > max_size:
        return hash_hex[:max_size]

    return hash_hex


def get_global_unique_hash(workspace: str, type: str, name: str) -> str:
    """
    Generate a unique hash for a combination of workspace, type, and name.

    Args:
        workspace: The workspace identifier
        type: The type identifier
        name: The name identifier

    Returns:
        A unique alphanumeric hash string of maximum length 48
    """
    global_unique_name = f"{workspace}-{type}-{name}"
    hash = get_alphanumeric_limited_hash(global_unique_name, 48)
    return hash


def pluralize(type_str: str) -> str:
    """
    Convert a string to its plural form following English pluralization rules.

    Args:
        type_str: The input string to pluralize

    Returns:
        The pluralized form of the input string
    """
    word = type_str.lower()

    # Words ending in s, ss, sh, ch, x, z - add 'es'
    if (
        word.endswith("s")
        or word.endswith("ss")
        or word.endswith("sh")
        or word.endswith("ch")
        or word.endswith("x")
        or word.endswith("z")
    ):
        return type_str + "es"

    # Words ending in consonant + y - change y to ies
    if word.endswith("y") and len(word) > 1:
        before_y = word[-2]
        if before_y not in "aeiou":
            return type_str[:-1] + "ies"

    # Words ending in f or fe - change to ves
    if word.endswith("f"):
        return type_str[:-1] + "ves"
    if word.endswith("fe"):
        return type_str[:-2] + "ves"

    # Words ending in consonant + o - add 'es'
    if word.endswith("o") and len(word) > 1:
        before_o = word[-2]
        if before_o not in "aeiou":
            return type_str + "es"

    # Default case - just add 's'
    return type_str + "s"


def get_forced_url(type_str: str, name: str) -> str | None:
    """
    Check for forced URLs in environment variables using both plural and singular forms.

    Args:
        type_str: The type identifier
        name: The name identifier

    Returns:
        The forced URL if found in environment variables, None otherwise
    """
    plural_type = pluralize(type_str)
    env_var = name.replace("-", "_").upper()

    # BL_FUNCTIONS_NAME_URL (plural form)
    plural_env_key = f"BL_{plural_type.upper()}_{env_var}_URL"
    if env[plural_env_key] is not None:
        return env[plural_env_key]

    # BL_FUNCTION_NAME_URL (singular form)
    singular_env_key = f"BL_{type_str.upper()}_{env_var}_URL"
    if env[singular_env_key] is not None:
        return env[singular_env_key]

    return None



from __future__ import annotations

import json
import logging
from typing import Any, Callable, Dict

import httpx

from ...client.models import Sandbox
from ..types import SandboxCreateConfiguration
from .sandbox import SandboxInstance


class CodeInterpreter(SandboxInstance):
    logger = logging.getLogger(__name__)
    DEFAULT_IMAGE = "blaxel/jupyter-server"
    DEFAULT_PORTS = [
        {"name": "jupyter", "target": 8888, "protocol": "HTTP"},
    ]
    DEFAULT_LIFECYCLE = {"expirationPolicies": [{"type": "ttl-idle", "value": "30m", "action": "delete"}]}

    @classmethod
    async def get(cls, sandbox_name: str) -> CodeInterpreter:
        base = await SandboxInstance.get(sandbox_name)
        return cls(
            sandbox=base.sandbox,
            force_url=base.config.force_url,
            headers=base.config.headers,
            params=base.config.params,
        )

    @classmethod
    async def create(
        cls,
        sandbox: Sandbox | SandboxCreateConfiguration | Dict[str, Any] | None = None,
        safe: bool = True,
    ) -> CodeInterpreter:
        """
        Create a sandbox instance using the jupyter-server image.
        Constraints:
          - Image is forced to blaxel/jupyter-server
          - Ports are fixed to 8888 (non-configurable)
          - Lifecycle defaults to 30m
          - No ttl, volumes, expires, snapshot_enabled
        """
        payload: Dict[str, Any] = {
            "image": cls.DEFAULT_IMAGE,
            "ports": cls.DEFAULT_PORTS,
            "lifecycle": cls.DEFAULT_LIFECYCLE,
        }

        # Whitelist a minimal set of fields that can be propagated from input
        allowed_copy_keys = {"name", "envs", "memory", "region", "headers"}

        if isinstance(sandbox, dict):
            for k in allowed_copy_keys:
                if k in sandbox and sandbox[k] is not None:
                    payload[k] = sandbox[k]
        elif isinstance(sandbox, SandboxCreateConfiguration):
            if getattr(sandbox, "name", None):
                payload["name"] = sandbox.name
            if getattr(sandbox, "envs", None):
                payload["envs"] = sandbox.envs
            if getattr(sandbox, "memory", None):
                payload["memory"] = sandbox.memory
            if getattr(sandbox, "region", None):
                payload["region"] = sandbox.region
        elif isinstance(sandbox, Sandbox):
            # Extract a few basics if available
            if sandbox.metadata and getattr(sandbox.metadata, "name", None):
                payload["name"] = sandbox.metadata.name
            if sandbox.spec and sandbox.spec.runtime:
                if getattr(sandbox.spec.runtime, "envs", None):
                    payload["envs"] = sandbox.spec.runtime.envs
                if getattr(sandbox.spec.runtime, "memory", None):
                    payload["memory"] = sandbox.spec.runtime.memory
            if sandbox.spec and getattr(sandbox.spec, "region", None):
                payload["region"] = sandbox.spec.region

        base_instance = await SandboxInstance.create(payload, safe=safe)
        return cls(
            sandbox=base_instance.sandbox,
            force_url=base_instance.config.force_url,
            headers=base_instance.config.headers,
            params=base_instance.config.params,
        )

    # Minimal internal types to satisfy API surface
    class OutputMessage:
        def __init__(self, text: str, timestamp: float | None, is_stderr: bool):
            self.text = text
            self.timestamp = timestamp
            self.is_stderr = is_stderr

    class Result:
        def __init__(self, **kwargs: Any):
            for k, v in kwargs.items():
                setattr(self, k, v)

    class ExecutionError:
        def __init__(self, name: str, value: Any, traceback: Any):
            self.name = name
            self.value = value
            self.traceback = traceback

    class Logs:
        def __init__(self):
            self.stdout: list[str] = []
            self.stderr: list[str] = []

    class Execution:
        def __init__(self):
            self.results: list[CodeInterpreter.Result] = []
            self.logs: CodeInterpreter.Logs = CodeInterpreter.Logs()
            self.error: CodeInterpreter.ExecutionError | None = None
            self.execution_count: int | None = None

    class Context:
        def __init__(self, id: str):
            self.id = id

        @classmethod
        def from_json(cls, data: Dict[str, Any]) -> CodeInterpreter.Context:
            return cls(id=str(data.get("id") or data.get("context_id") or ""))

    @property
    def _jupyter_url(self) -> str:
        # Use the same base as other async actions
        return self.process.url

    def _parse_output(
        self,
        execution: CodeInterpreter.Execution,
        output: str,
        on_stdout: Callable[[Any], Any] | None = None,
        on_stderr: Callable[[Any], Any] | None = None,
        on_result: Callable[[Any], Any] | None = None,
        on_error: Callable[[Any], Any] | None = None,
    ) -> Any | None:
        data = json.loads(output)
        data_type = data.pop("type")

        if data_type == "result":
            result = CodeInterpreter.Result(**data)
            execution.results.append(result)
            if on_result:
                return on_result(result)
        elif data_type == "stdout":
            execution.logs.stdout.append(data["text"])
            if on_stdout:
                return on_stdout(CodeInterpreter.OutputMessage(data["text"], data.get("timestamp"), False))
        elif data_type == "stderr":
            execution.logs.stderr.append(data["text"])
            if on_stderr:
                return on_stderr(CodeInterpreter.OutputMessage(data["text"], data.get("timestamp"), True))
        elif data_type == "error":
            execution.error = CodeInterpreter.ExecutionError(
                data.get("name", ""), data.get("value"), data.get("traceback")
            )
            if on_error:
                return on_error(execution.error)
        elif data_type == "number_of_executions":
            execution.execution_count = data.get("execution_count")

        return None

    async def run_code(
        self,
        code: str,
        language: str | None = None,
        context: CodeInterpreter.Context | None = None,
        on_stdout: Callable[[Any], None] | None = None,
        on_stderr: Callable[[Any], None] | None = None,
        on_result: Callable[[Any], None] | None = None,
        on_error: Callable[[Any], None] | None = None,
        envs: Dict[str, str] | None = None,
        timeout: float | None = None,
        request_timeout: float | None = None,
    ) -> CodeInterpreter.Execution:
        # Defaults: treat 0 as no read timeout
        DEFAULT_TIMEOUT = 60.0
        if language and context:
            raise ValueError("You can provide context or language, but not both at the same time.")

        read_timeout = None if timeout == 0 else (timeout or DEFAULT_TIMEOUT)
        connect_timeout = request_timeout or 10.0
        write_timeout = request_timeout or 10.0
        pool_timeout = request_timeout or 10.0

        context_id = context.id if context else None

        body = {
            "code": code,
            "context_id": context_id,
            "language": language,
            "env_vars": envs,
        }

        execution = CodeInterpreter.Execution()

        async with self.process.get_client() as client:
            timeout_cfg = httpx.Timeout(
                connect=connect_timeout, read=read_timeout, write=write_timeout, pool=pool_timeout
            )
            async with client.stream(
                "POST",
                "/port/8888/execute",
                json=body,
                timeout=timeout_cfg,
            ) as response:
                if response.status_code >= 400:
                    try:
                        body_text = await response.aread()
                        body_text = body_text.decode(errors="ignore")
                    except Exception:
                        body_text = "<unavailable>"
                    req = getattr(response, "request", None)
                    method = getattr(req, "method", "UNKNOWN") if req else "UNKNOWN"
                    url = str(getattr(req, "url", "UNKNOWN")) if req else "UNKNOWN"
                    reason = getattr(response, "reason_phrase", "")
                    details = (
                        "Execution failed\n"
                        f"- method: {method}\n- url: {url}\n- status: {response.status_code} {reason}\n"
                        f"- response-headers: {dict(response.headers)}\n- body:\n{body_text}"
                    )
                    self.logger.debug(details)
                    raise RuntimeError(details)

                async for line in response.aiter_lines():
                    if not line:
                        continue
                    try:
                        decoded = line
                    except Exception:
                        decoded = str(line)
                    try:
                        self._parse_output(
                            execution,
                            decoded,
                            on_stdout=on_stdout,
                            on_stderr=on_stderr,
                            on_result=on_result,
                            on_error=on_error,
                        )
                    except json.JSONDecodeError:
                        # Fallback: treat as stdout text-only message
                        execution.logs.stdout.append(decoded)
                        if on_stdout:
                            on_stdout(CodeInterpreter.OutputMessage(decoded, None, False))

        return execution

    async def create_code_context(
        self,
        cwd: str | None = None,
        language: str | None = None,
        request_timeout: float | None = None,
    ) -> CodeInterpreter.Context:
        data: Dict[str, Any] = {}
        if language:
            data["language"] = language
        if cwd:
            data["cwd"] = cwd

        async with self.process.get_client() as client:
            response = await client.post(
                "/port/8888/contexts",
                json=data,
                timeout=request_timeout or 10.0,
            )
            if response.status_code >= 400:
                try:
                    body_text = response.text
                except Exception:
                    body_text = "<unavailable>"
                method = getattr(response.request, "method", "UNKNOWN")
                url = str(getattr(response.request, "url", "UNKNOWN"))
                reason = getattr(response, "reason_phrase", "")
                details = (
                    "Create context failed\n"
                    f"- method: {method}\n- url: {url}\n- status: {response.status_code} {reason}\n"
                    f"- response-headers: {dict(response.headers)}\n- body:\n{body_text}"
                )
                self.logger.debug(details)
                raise RuntimeError(details)
            data = response.json()
            return CodeInterpreter.Context.from_json(data)





import uuid
import warnings
from typing import Any, Callable, Dict, List, Sequence, Union

import cohere
from cohere.types import ToolCall
from llama_index.core.base.llms.types import (
    ChatMessage,
    ChatResponse,
    ChatResponseAsyncGen,
    ChatResponseGen,
    CompletionResponse,
    CompletionResponseAsyncGen,
    CompletionResponseGen,
    LLMMetadata,
    MessageRole,
)
from llama_index.core.bridge.pydantic import Field, PrivateAttr
from llama_index.core.callbacks import CallbackManager
from llama_index.core.llms.callbacks import llm_chat_callback, llm_completion_callback
from llama_index.core.llms.function_calling import FunctionCallingLLM
from llama_index.core.llms.llm import ToolSelection
from llama_index.core.tools.types import BaseTool
from llama_index.core.types import BaseOutputParser, PydanticProgramMode
from llama_index.llms.cohere.utils import (
    CHAT_MODELS,
    _get_message_cohere_format,
    _message_to_cohere_tool_results,
    _messages_to_cohere_tool_results_curr_chat_turn,
    acompletion_with_retry,
    cohere_modelname_to_contextsize,
    completion_with_retry,
    format_to_cohere_tools,
    is_cohere_function_calling_model,
    remove_documents_from_messages,
)


class Cohere(FunctionCallingLLM):
    """Cohere LLM.

    Examples:
        `pip install llama-index-llms-cohere`

        ```python
        from llama_index.llms.cohere import Cohere

        llm = Cohere(model="command", api_key=api_key)
        resp = llm.complete("Paul Graham is ")
        ```
    """

    model: str = Field(description="The cohere model to use.")
    temperature: float | None = Field(
        description="The temperature to use for sampling.", default=None
    )
    max_retries: int = Field(default=10, description="The maximum number of API retries.")
    additional_kwargs: Dict[str, Any] = Field(
        default_factory=dict, description="Additional kwargs for the Cohere API."
    )
    max_tokens: int = Field(description="The maximum number of tokens to generate.")

    _client: Any = PrivateAttr()
    _aclient: Any = PrivateAttr()

    def __init__(
        self,
        model: str = "command-r",
        temperature: float | None = None,
        max_tokens: int | None = 8192,
        timeout: float | None = None,
        max_retries: int = 10,
        api_key: str | None = None,
        api_base: str | None = None,
        additional_kwargs: Dict[str, Any] | None = None,
        callback_manager: CallbackManager | None = None,
        system_prompt: str | None = None,
        messages_to_prompt: Callable[[Sequence[ChatMessage]], str] | None = None,
        completion_to_prompt: Callable[[str], str] | None = None,
        pydantic_program_mode: PydanticProgramMode = PydanticProgramMode.DEFAULT,
        output_parser: BaseOutputParser | None = None,
    ) -> None:
        additional_kwargs = additional_kwargs or {}
        callback_manager = callback_manager or CallbackManager([])

        super().__init__(
            temperature=temperature,
            additional_kwargs=additional_kwargs,
            timeout=timeout,
            max_retries=max_retries,
            model=model,
            callback_manager=callback_manager,
            max_tokens=max_tokens,
            system_prompt=system_prompt,
            messages_to_prompt=messages_to_prompt,
            completion_to_prompt=completion_to_prompt,
            pydantic_program_mode=pydantic_program_mode,
            output_parser=output_parser,
        )
        self._client = cohere.Client(api_key, client_name="llama_index", base_url=api_base)
        self._aclient = cohere.AsyncClient(api_key, client_name="llama_index", base_url=api_base)

    @classmethod
    def class_name(cls) -> str:
        """Get class name."""
        return "Cohere_LLM"

    @property
    def metadata(self) -> LLMMetadata:
        return LLMMetadata(
            context_window=cohere_modelname_to_contextsize(self.model),
            num_output=self.max_tokens,
            is_chat_model=True,
            model_name=self.model,
            system_role=MessageRole.CHATBOT,
            is_function_calling_model=is_cohere_function_calling_model(self.model),
        )

    @property
    def _model_kwargs(self) -> Dict[str, Any]:
        base_kwargs = {
            "model": self.model,
            "temperature": self.temperature,
        }
        return {
            **base_kwargs,
            **self.additional_kwargs,
        }

    def _get_all_kwargs(self, **kwargs: Any) -> Dict[str, Any]:
        return {
            **self._model_kwargs,
            **kwargs,
        }

    def _prepare_chat_with_tools(
        self,
        tools: List["BaseTool"],
        user_msg: Union[str, ChatMessage] | None = None,
        chat_history: List[ChatMessage] | None = None,
        verbose: bool = False,
        allow_parallel_tool_calls: bool = False,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Prepare the chat with tools."""
        chat_history = chat_history or []

        if isinstance(user_msg, str):
            user_msg = ChatMessage(role=MessageRole.USER, content=user_msg)

        if user_msg is not None:
            chat_history.append(user_msg)

        tools_cohere_format = format_to_cohere_tools(tools)
        return {
            "messages": chat_history,
            "tools": tools_cohere_format or [],
            **kwargs,
        }

    def get_tool_calls_from_response(
        self,
        response: "ChatResponse",
        error_on_no_tool_call: bool = False,
    ) -> List[ToolSelection]:
        """Predict and call the tool."""
        tool_calls: List[ToolCall] = response.message.additional_kwargs.get("tool_calls", []) or []

        if len(tool_calls) < 1 and error_on_no_tool_call:
            raise ValueError(
                f"Expected at least one tool call, but got {len(tool_calls)} tool calls."
            )

        tool_selections = []
        for tool_call in tool_calls:
            if not isinstance(tool_call, ToolCall):
                raise ValueError("Invalid tool_call object")
            tool_selections.append(
                ToolSelection(
                    tool_id=uuid.uuid4().hex[:],
                    tool_name=tool_call.name,
                    tool_kwargs=tool_call.parameters,
                )
            )

        return tool_selections

    def get_cohere_chat_request(
        self,
        messages: List[ChatMessage],
        *,
        connectors: List[Dict[str, str]] | None = None,
        stop_sequences: List[str] | None = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Get the request for the Cohere chat API.

        Args:
            messages: The messages.
            connectors: The connectors.
            **kwargs: The keyword arguments.

        Returns:
            The request for the Cohere chat API.
        """
        additional_kwargs = messages[-1].additional_kwargs
        documents = additional_kwargs.pop("documents", [])
        # cohere SDK will fail loudly if both connectors and documents are provided
        if documents and len(documents) > 0:
            raise ValueError(
                "Received documents both as a keyword argument and as an prompt additional keyword argument. Please choose only one option."
            )

        messages, documents = remove_documents_from_messages(messages)

        tool_results: List[Dict[str, Any]] | None = (
            _messages_to_cohere_tool_results_curr_chat_turn(messages) or kwargs.get("tool_results")
        )
        if not tool_results:
            tool_results = None

        chat_history = []
        temp_tool_results = []
        # if force_single_step is set to False, then only message is empty in request if there is tool call
        if not kwargs.get("force_single_step"):
            for i, message in enumerate(messages[:-1]):
                # If there are multiple tool messages, then we need to aggregate them into one single tool message to pass into chat history
                if message.role == MessageRole.TOOL:
                    temp_tool_results += _message_to_cohere_tool_results(messages, i)

                    if (i == len(messages) - 1) or messages[i + 1].role != MessageRole.TOOL:
                        cohere_message = _get_message_cohere_format(message, temp_tool_results)
                        chat_history.append(cohere_message)
                        temp_tool_results = []
                else:
                    chat_history.append(_get_message_cohere_format(message, None))

            message_str = "" if tool_results else messages[-1].content

        else:
            message_str = ""
            # if force_single_step is set to True, then message is the last human message in the conversation
            for message in messages[:-1]:
                if message.role in (
                    MessageRole.CHATBOT,
                    MessageRole.ASSISTANT,
                ) and message.additional_kwargs.get("tool_calls"):
                    continue

                # If there are multiple tool messages, then we need to aggregate them into one single tool message to pass into chat history
                if message.role == MessageRole.TOOL:
                    temp_tool_results += _message_to_cohere_tool_results(messages, i)

                    if (i == len(messages) - 1) or messages[i + 1].role != MessageRole.TOOL:
                        cohere_message = _get_message_cohere_format(message, temp_tool_results)
                        chat_history.append(cohere_message)
                        temp_tool_results = []
                else:
                    chat_history.append(_get_message_cohere_format(message, None))
            # Add the last human message in the conversation to the message string
            for message in messages[::-1]:
                if (message.role == MessageRole.USER) and (message.content):
                    message_str = message.content
                    break

        req = {
            "message": message_str,
            "chat_history": chat_history,
            "tool_results": tool_results,
            "documents": documents,
            "connectors": connectors,
            "stop_sequences": stop_sequences,
            **kwargs,
        }
        return {k: v for k, v in req.items() if v is not None}

    @llm_chat_callback()
    def chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        all_kwargs = self._get_all_kwargs(**kwargs)

        chat_request = self.get_cohere_chat_request(messages=messages, **all_kwargs)

        if all_kwargs["model"] not in CHAT_MODELS:
            raise ValueError(f"{all_kwargs['model']} not supported for chat")

        if "stream" in all_kwargs:
            warnings.warn(
                "Parameter `stream` is not supported by the `chat` method."
                "Use the `stream_chat` method instead"
            )

        response = completion_with_retry(
            client=self._client, max_retries=self.max_retries, chat=True, **chat_request
        )
        if not isinstance(response, cohere.NonStreamedChatResponse):
            tool_calls = response.get("tool_calls")
            content = response.get("text")
            response_raw = response

        else:
            tool_calls = response.tool_calls
            content = response.text
            response_raw = response.__dict__

        return ChatResponse(
            message=ChatMessage(
                role=MessageRole.ASSISTANT,
                content=content,
                additional_kwargs={"tool_calls": tool_calls},
            ),
            raw=response_raw,
        )

    @llm_completion_callback()
    def complete(self, prompt: str, formatted: bool = False, **kwargs: Any) -> CompletionResponse:
        all_kwargs = self._get_all_kwargs(**kwargs)
        if "stream" in all_kwargs:
            warnings.warn(
                "Parameter `stream` is not supported by the `chat` method."
                "Use the `stream_chat` method instead"
            )

        response = completion_with_retry(
            client=self._client,
            max_retries=self.max_retries,
            chat=False,
            prompt=prompt,
            **all_kwargs,
        )

        return CompletionResponse(
            text=response.generations[0].text,
            raw=response.__dict__,
        )

    @llm_chat_callback()
    def stream_chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponseGen:
        all_kwargs = self._get_all_kwargs(**kwargs)
        all_kwargs["stream"] = True
        if all_kwargs["model"] not in CHAT_MODELS:
            raise ValueError(f"{all_kwargs['model']} not supported for chat")

        chat_request = self.get_cohere_chat_request(messages=messages, **all_kwargs)

        response = completion_with_retry(
            client=self._client, max_retries=self.max_retries, chat=True, **chat_request
        )

        def gen() -> ChatResponseGen:
            content = ""
            role = MessageRole.ASSISTANT
            for r in response:
                if "text" in r.__dict__:
                    content_delta = r.text
                else:
                    content_delta = ""
                content += content_delta
                yield ChatResponse(
                    message=ChatMessage(role=role, content=content),
                    delta=content_delta,
                    raw=r.__dict__,
                )

        return gen()

    @llm_completion_callback()
    def stream_complete(
        self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponseGen:
        all_kwargs = self._get_all_kwargs(**kwargs)
        all_kwargs["stream"] = True

        response = completion_with_retry(
            client=self._client,
            max_retries=self.max_retries,
            chat=False,
            prompt=prompt,
            **all_kwargs,
        )

        def gen() -> CompletionResponseGen:
            content = ""
            for r in response:
                content_delta = r.text
                content += content_delta
                yield CompletionResponse(text=content, delta=content_delta, raw=r._asdict())

        return gen()

    @llm_chat_callback()
    async def achat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        all_kwargs = self._get_all_kwargs(**kwargs)
        if all_kwargs["model"] not in CHAT_MODELS:
            raise ValueError(f"{all_kwargs['model']} not supported for chat")
        if "stream" in all_kwargs:
            warnings.warn(
                "Parameter `stream` is not supported by the `chat` method."
                "Use the `stream_chat` method instead"
            )

        chat_request = self.get_cohere_chat_request(messages=messages, **all_kwargs)

        response = await acompletion_with_retry(
            aclient=self._aclient,
            max_retries=self.max_retries,
            chat=True,
            **chat_request,
        )

        if not isinstance(response, cohere.NonStreamedChatResponse):
            tool_calls = response.get("tool_calls")
            content = response.get("text")
            response_raw = response

        else:
            tool_calls = response.tool_calls
            content = response.text
            response_raw = response.__dict__

        if not isinstance(response, cohere.NonStreamedChatResponse):
            tool_calls = response.get("tool_calls")
            content = response.get("text")
            response_raw = response

        else:
            tool_calls = response.tool_calls
            content = response.text
            response_raw = response.__dict__

        return ChatResponse(
            message=ChatMessage(
                role=MessageRole.ASSISTANT,
                content=content,
                additional_kwargs={"tool_calls": tool_calls},
            ),
            raw=response_raw,
        )

    @llm_completion_callback()
    async def acomplete(
        self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponse:
        all_kwargs = self._get_all_kwargs(**kwargs)
        if "stream" in all_kwargs:
            warnings.warn(
                "Parameter `stream` is not supported by the `chat` method."
                "Use the `stream_chat` method instead"
            )

        response = await acompletion_with_retry(
            aclient=self._aclient,
            max_retries=self.max_retries,
            chat=False,
            prompt=prompt,
            **all_kwargs,
        )

        return CompletionResponse(
            text=response.generations[0].text,
            raw=response.__dict__,
        )

    @llm_chat_callback()
    async def astream_chat(
        self, messages: Sequence[ChatMessage], **kwargs: Any
    ) -> ChatResponseAsyncGen:
        all_kwargs = self._get_all_kwargs(**kwargs)
        all_kwargs["stream"] = True
        if all_kwargs["model"] not in CHAT_MODELS:
            raise ValueError(f"{all_kwargs['model']} not supported for chat")

        chat_request = self.get_cohere_chat_request(messages, **all_kwargs)

        response = completion_with_retry(
            client=self._client, max_retries=self.max_retries, chat=True, **chat_request
        )

        async def gen() -> ChatResponseAsyncGen:
            content = ""
            role = MessageRole.ASSISTANT
            async for r in response:
                if "text" in r.__dict__:
                    content_delta = r.text
                else:
                    content_delta = ""
                content += content_delta
                yield ChatResponse(
                    message=ChatMessage(role=role, content=content),
                    delta=content_delta,
                    raw=r.__dict__,
                )

        return gen()

    @llm_completion_callback()
    async def astream_complete(
        self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponseAsyncGen:
        all_kwargs = self._get_all_kwargs(**kwargs)
        all_kwargs["stream"] = True

        response = await acompletion_with_retry(
            aclient=self._aclient,
            max_retries=self.max_retries,
            chat=False,
            prompt=prompt,
            **all_kwargs,
        )

        async def gen() -> CompletionResponseAsyncGen:
            content = ""
            async for r in response:
                content_delta = r.text
                content += content_delta
                yield CompletionResponse(text=content, delta=content_delta, raw=r._asdict())

        return gen()



import asyncio
import io
import json
from pathlib import Path
from typing import Any, Callable, Dict, List, Union

import httpx

from ...common.settings import settings
from ..client.models import Directory, FileRequest, SuccessResponse
from ..types import (
    CopyResponse,
    SandboxConfiguration,
    SandboxFilesystemFile,
    WatchEvent,
)
from .action import SandboxAction

# Multipart upload constants
MULTIPART_THRESHOLD = 5 * 1024 * 1024  # 5MB
CHUNK_SIZE = 5 * 1024 * 1024  # 5MB per part
MAX_PARALLEL_UPLOADS = 3  # Number of parallel part uploads


class SandboxFileSystem(SandboxAction):
    def __init__(self, sandbox_config: SandboxConfiguration, process=None):
        super().__init__(sandbox_config)
        self.process = process

    async def mkdir(self, path: str, permissions: str = "0755") -> SuccessResponse:
        path = self.format_path(path)
        body = FileRequest(is_directory=True, permissions=permissions)

        async with self.get_client() as client_instance:
            response = await client_instance.put(f"/filesystem/{path}", json=body.to_dict())
            self.handle_response_error(response)
            return SuccessResponse.from_dict(response.json())

    async def write(self, path: str, content: str) -> SuccessResponse:
        path = self.format_path(path)

        # Calculate content size in bytes
        content_size = len(content.encode('utf-8'))

        # Use multipart upload for large files
        if content_size > MULTIPART_THRESHOLD:
            content_bytes = content.encode('utf-8')
            return await self._upload_with_multipart(path, content_bytes, "0644")

        # Use regular upload for small files
        body = FileRequest(content=content)

        async with self.get_client() as client_instance:
            response = await client_instance.put(f"/filesystem/{path}", json=body.to_dict())
            self.handle_response_error(response)
            return SuccessResponse.from_dict(response.json())

    async def write_binary(self, path: str, content: Union[bytes, bytearray, str]) -> SuccessResponse:
        """Write binary content to a file.

        Args:
            path: The path in the sandbox to write to
            content: Binary content as bytes, bytearray, or string path to a local file

        Returns:
            SuccessResponse indicating success
        """
        path = self.format_path(path)

        # If content is a string, treat it as a file path and read it
        if isinstance(content, str):
            local_path = Path(content)
            content = local_path.read_bytes()
        # Convert bytearray to bytes if necessary
        elif isinstance(content, bytearray):
            content = bytes(content)

        # Use multipart upload for large files
        if len(content) > MULTIPART_THRESHOLD:
            return await self._upload_with_multipart(path, content, "0644")

        # Use regular upload for small files
        # Wrap binary content in BytesIO to provide file-like interface
        binary_file = io.BytesIO(content)

        # Prepare multipart form data
        files = {
            "file": ("binary-file.bin", binary_file, "application/octet-stream"),
        }
        data = {"permissions": "0644", "path": path}

        # Use the fixed get_client method
        url = f"{self.url}/filesystem/{path}"
        headers = {**settings.headers, **self.sandbox_config.headers}

        async with self.get_client() as client_instance:
            response = await client_instance.put(url, files=files, data=data, headers=headers)

            if not response.is_success:
                raise Exception(f"Failed to write binary: {response.status_code} {response.text}")

            return SuccessResponse.from_dict(response.json())

    async def write_tree(
        self,
        files: List[Union[SandboxFilesystemFile, Dict[str, Any]]],
        destination_path: str | None = None,
    ) -> Directory:
        """Write multiple files in a tree structure."""
        files_dict = {}
        for file in files:
            if isinstance(file, dict):
                file = SandboxFilesystemFile.from_dict(file)
            files_dict[file.path] = file.content

        path = destination_path or ""

        async with self.get_client() as client_instance:
            response = await client_instance.put(
                f"/filesystem/tree/{path}",
                json={"files": files_dict},
                headers={"Content-Type": "application/json"},
            )
            self.handle_response_error(response)
            return Directory.from_dict(response.json())

    async def read(self, path: str) -> str:
        path = self.format_path(path)

        async with self.get_client() as client_instance:
            response = await client_instance.get(f"/filesystem/{path}")
            self.handle_response_error(response)

            data = response.json()
            if "content" in data:
                return data["content"]
            raise Exception("Unsupported file type")

    async def read_binary(self, path: str) -> bytes:
        """Read binary content from a file.

        Args:
            path: The path in the sandbox to read from

        Returns:
            Binary content as bytes
        """
        path = self.format_path(path)

        url = f"{self.url}/filesystem/{path}"
        headers = {
            **settings.headers,
            **self.sandbox_config.headers,
            "Accept": "application/octet-stream",
        }

        async with self.get_client() as client_instance:
            response = await client_instance.get(url, headers=headers)
            self.handle_response_error(response)
            return response.content

    async def download(self, src: str, destination_path: str, mode: int = 0o644) -> None:
        """Download a file from the sandbox to the local filesystem.

        Args:
            src: The path in the sandbox to download from
            destination_path: The local path to save to
            mode: File permissions mode (default: 0o644)
        """
        content = await self.read_binary(src)
        local_path = Path(destination_path)
        local_path.write_bytes(content)
        local_path.chmod(mode)

    async def rm(self, path: str, recursive: bool = False) -> SuccessResponse:
        path = self.format_path(path)

        async with self.get_client() as client_instance:
            params = {"recursive": "true"} if recursive else {}
            response = await client_instance.delete(f"/filesystem/{path}", params=params)
            self.handle_response_error(response)
            return SuccessResponse.from_dict(response.json())

    async def ls(self, path: str) -> Directory:
        path = self.format_path(path)

        async with self.get_client() as client_instance:
            response = await client_instance.get(f"/filesystem/{path}")
            self.handle_response_error(response)

            data = response.json()
            if not ("files" in data or "subdirectories" in data):
                raise Exception('{"error": "Directory not found"}')
            return Directory.from_dict(data)

    async def find(
        self,
        path: str,
        type: str | None = None,
        patterns: List[str] | None = None,
        max_results: int | None = None,
        exclude_dirs: List[str] | None = None,
        exclude_hidden: bool | None = None,
    ):
        """Find files and directories.

        Args:
            path: Path to search in
            type: Type of search ('file' or 'directory')
            patterns: File patterns to include (e.g., ['*.py', '*.json'])
            max_results: Maximum number of results to return
            exclude_dirs: Directory names to skip
            exclude_hidden: Exclude hidden files and directories

        Returns:
            FindResponse with matching files/directories
        """
        path = self.format_path(path)
        
        params = {}
        if type is not None:
            params['type'] = type
        if patterns is not None and len(patterns) > 0:
            params['patterns'] = ','.join(patterns)
        if max_results is not None:
            params['maxResults'] = max_results
        if exclude_dirs is not None and len(exclude_dirs) > 0:
            params['excludeDirs'] = ','.join(exclude_dirs)
        if exclude_hidden is not None:
            params['excludeHidden'] = exclude_hidden

        url = f"{self.url}/filesystem-find/{path}"
        headers = {**settings.headers, **self.sandbox_config.headers}

        async with self.get_client() as client_instance:
            response = await client_instance.get(url, params=params, headers=headers)
            self.handle_response_error(response)
            
            from ..client.models.find_response import FindResponse
            return FindResponse.from_dict(response.json())

    async def grep(
        self,
        query: str,
        path: str = "/",
        case_sensitive: bool | None = None,
        context_lines: int | None = None,
        max_results: int | None = None,
        file_pattern: str | None = None,
        exclude_dirs: List[str] | None = None,
    ):
        """Search for text content inside files using ripgrep.

        Args:
            query: Text to search for
            path: Directory path to search in
            case_sensitive: Case sensitive search (default: false)
            context_lines: Number of context lines to include (default: 0)
            max_results: Maximum number of results to return (default: 100)
            file_pattern: File pattern to include (e.g., '*.py')
            exclude_dirs: Directory names to skip

        Returns:
            ContentSearchResponse with matching lines
        """
        path = self.format_path(path)
        
        params = {'query': query}
        if case_sensitive is not None:
            params['caseSensitive'] = case_sensitive
        if context_lines is not None:
            params['contextLines'] = context_lines
        if max_results is not None:
            params['maxResults'] = max_results
        if file_pattern is not None:
            params['filePattern'] = file_pattern
        if exclude_dirs is not None and len(exclude_dirs) > 0:
            params['excludeDirs'] = ','.join(exclude_dirs)

        url = f"{self.url}/filesystem-content-search/{path}"
        headers = {**settings.headers, **self.sandbox_config.headers}

        async with self.get_client() as client_instance:
            response = await client_instance.get(url, params=params, headers=headers)
            self.handle_response_error(response)
            
            from ..client.models.content_search_response import ContentSearchResponse
            return ContentSearchResponse.from_dict(response.json())

    async def cp(self, source: str, destination: str, max_wait: int = 180000) -> CopyResponse:
        """Copy files or directories using the cp command.

        Args:
            source: Source path
            destination: Destination path
            max_wait: Maximum time to wait for the copy operation in milliseconds (default: 180000)
        """
        if not self.process:
            raise Exception("Process instance not available. Cannot execute cp command.")

        # Execute cp -r command
        process = await self.process.exec({
            "command": f"cp -r {source} {destination}"
        })

        # Wait for process to complete
        process = await self.process.wait(process.pid, max_wait=max_wait, interval=100)

        # Check if process failed
        if process.status == "failed":
            logs = process.logs if hasattr(process, "logs") else "Unknown error"
            raise Exception(f"Could not copy {source} to {destination} cause: {logs}")

        return CopyResponse(
            message="Files copied",
            source=source,
            destination=destination
        )

    def watch(
        self,
        path: str,
        callback: Callable[[WatchEvent], None],
        options: Dict[str, Any] | None = None,
    ) -> Dict[str, Callable]:
        """Watch for file system changes."""
        path = self.format_path(path)
        closed = False

        if options is None:
            options = {}

        async def start_watching():
            nonlocal closed
            params = {}
            if options.get("ignore"):
                params["ignore"] = ",".join(options["ignore"])

            url = f"{self.url}/watch/filesystem/{path}"
            headers = {**settings.headers, **self.sandbox_config.headers}
            async with httpx.AsyncClient() as client_instance:
                async with client_instance.stream(
                    "GET", url, params=params, headers=headers
                ) as response:
                    if not response.is_success:
                        raise Exception(f"Failed to start watching: {response.status_code}")
                    buffer = ""
                    async for chunk in response.aiter_text():
                        if closed:
                            break

                        buffer += chunk
                        lines = buffer.split("\n")
                        buffer = lines.pop()  # Keep incomplete line in buffer

                        for line in lines:
                            line = line.strip()
                            if not line:
                                continue

                            # Skip keepalive messages
                            if line.startswith("[keepalive]"):
                                continue

                            try:
                                file_event_data = json.loads(line)
                                file_event = WatchEvent(
                                    op=file_event_data.get("op", ""),
                                    path=file_event_data.get("path", ""),
                                    name=file_event_data.get("name", ""),
                                    content=file_event_data.get("content"),
                                )

                                if options.get("with_content") and file_event.op in [
                                    "CREATE",
                                    "WRITE",
                                ]:
                                    try:
                                        file_path = file_event.path
                                        if file_path.endswith("/"):
                                            file_path = file_path + file_event.name
                                        else:
                                            file_path = file_path + "/" + file_event.name

                                        content = await self.read(file_path)
                                        file_event.content = content
                                    except:
                                        file_event.content = None

                                await callback(file_event)
                            except json.JSONDecodeError:
                                continue
                            except Exception as e:
                                if options.get("on_error"):
                                    options["on_error"](e)

        # Start watching in the background
        task = asyncio.create_task(start_watching())

        def close():
            nonlocal closed
            closed = True
            task.cancel()

        return {"close": close}

    def format_path(self, path: str) -> str:
        """Format path for filesystem operations.

        Simplified to match TypeScript behavior - returns path as-is.
        """
        return path

    # Multipart upload helper methods
    async def _initiate_multipart_upload(self, path: str, permissions: str = "0644") -> Dict[str, Any]:
        """Initiate a multipart upload session."""
        path = self.format_path(path)
        url = f"{self.url}/filesystem-multipart/initiate/{path}"
        headers = {**settings.headers, **self.sandbox_config.headers}
        body = {"permissions": permissions}

        async with self.get_client() as client_instance:
            response = await client_instance.post(url, json=body, headers=headers)
            self.handle_response_error(response)
            return response.json()

    async def _upload_part(
        self, upload_id: str, part_number: int, data: bytes
    ) -> Dict[str, Any]:
        """Upload a single part of a multipart upload."""
        url = f"{self.url}/filesystem-multipart/{upload_id}/part"
        headers = {**settings.headers, **self.sandbox_config.headers}
        params = {"partNumber": part_number}

        # Prepare multipart form data with the file chunk
        files = {"file": ("part", io.BytesIO(data), "application/octet-stream")}

        async with self.get_client() as client_instance:
            response = await client_instance.put(
                url, files=files, params=params, headers=headers
            )
            self.handle_response_error(response)
            return response.json()

    async def _complete_multipart_upload(
        self, upload_id: str, parts: List[Dict[str, Any]]
    ) -> SuccessResponse:
        """Complete a multipart upload by assembling all parts."""
        url = f"{self.url}/filesystem-multipart/{upload_id}/complete"
        headers = {**settings.headers, **self.sandbox_config.headers}
        body = {"parts": parts}

        async with self.get_client() as client_instance:
            response = await client_instance.post(url, json=body, headers=headers)
            self.handle_response_error(response)
            return SuccessResponse.from_dict(response.json())

    async def _abort_multipart_upload(self, upload_id: str) -> None:
        """Abort a multipart upload and clean up all parts."""
        url = f"{self.url}/filesystem-multipart/{upload_id}/abort"
        headers = {**settings.headers, **self.sandbox_config.headers}

        async with self.get_client() as client_instance:
            response = await client_instance.delete(url, headers=headers)
            # Don't raise error if abort fails - we want to throw the original error
            if not response.is_success:
                print(f"Warning: Failed to abort multipart upload: {response.status_code}")

    async def _upload_with_multipart(
        self, path: str, data: bytes, permissions: str = "0644"
    ) -> SuccessResponse:
        """Upload a file using multipart upload for large files."""
        # Initiate multipart upload
        init_response = await self._initiate_multipart_upload(path, permissions)
        upload_id = init_response.get("uploadId")

        if not upload_id:
            raise Exception("Failed to get upload ID from initiate response")

        try:
            size = len(data)
            num_parts = (size + CHUNK_SIZE - 1) // CHUNK_SIZE  # Ceiling division
            parts: List[Dict[str, Any]] = []

            # Upload parts in batches for parallel processing
            for i in range(0, num_parts, MAX_PARALLEL_UPLOADS):
                batch_tasks = []

                for j in range(MAX_PARALLEL_UPLOADS):
                    if i + j >= num_parts:
                        break

                    part_number = i + j + 1
                    start = (part_number - 1) * CHUNK_SIZE
                    end = min(start + CHUNK_SIZE, size)
                    chunk = data[start:end]

                    batch_tasks.append(self._upload_part(upload_id, part_number, chunk))

                # Wait for batch to complete
                batch_results = await asyncio.gather(*batch_tasks)
                parts.extend(
                    [
                        {
                            "partNumber": r.get("partNumber"),
                            "etag": r.get("etag"),
                        }
                        for r in batch_results
                    ]
                )

            # Sort parts by partNumber to ensure correct order
            parts.sort(key=lambda p: p.get("partNumber", 0))

            # Complete the upload
            return await self._complete_multipart_upload(upload_id, parts)
        except Exception as error:
            # Abort the upload on failure
            try:
                await self._abort_multipart_upload(upload_id)
            except Exception as abort_error:
                # Log but don't throw - we want to throw the original error
                print(f"Failed to abort multipart upload: {abort_error}")
            raise error



from http import HTTPStatus
from typing import Any, Union

import httpx

from ... import errors
from ...client import Client
from ...models.apply_edit_request import ApplyEditRequest
from ...models.apply_edit_response import ApplyEditResponse
from ...models.error_response import ErrorResponse
from ...types import Response


def _get_kwargs(
    path: str,
    *,
    body: ApplyEditRequest,
) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    _kwargs: dict[str, Any] = {
        "method": "put",
        "url": f"/codegen/fastapply/{path}",
    }

    if type(body) is dict:
        _body = body
    else:
        _body = body.to_dict()

    _kwargs["json"] = _body
    headers["Content-Type"] = "application/json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(*, client: Client, response: httpx.Response) -> Union[ApplyEditResponse, ErrorResponse] | None:
    if response.status_code == 200:
        response_200 = ApplyEditResponse.from_dict(response.json())

        return response_200
    if response.status_code == 400:
        response_400 = ErrorResponse.from_dict(response.json())

        return response_400
    if response.status_code == 422:
        response_422 = ErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == 503:
        response_503 = ErrorResponse.from_dict(response.json())

        return response_503
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(*, client: Client, response: httpx.Response) -> Response[Union[ApplyEditResponse, ErrorResponse]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    path: str,
    *,
    client: Union[Client],
    body: ApplyEditRequest,
) -> Response[Union[ApplyEditResponse, ErrorResponse]]:
    r"""Apply code edit

     Uses the configured LLM provider (Relace or Morph) to apply a code edit to the original content.

    To use this endpoint as an agent tool, follow these guidelines:

    Use this tool to make an edit to an existing file. This will be read by a less intelligent model,
    which will quickly apply the edit. You should make it clear what the edit is, while also minimizing
    the unchanged code you write.

    When writing the edit, you should specify each edit in sequence, with the special comment \"// ...
    existing code ...\" to represent unchanged code in between edited lines.

    Example format:
    // ... existing code ...
    FIRST_EDIT
    // ... existing code ...
    SECOND_EDIT
    // ... existing code ...
    THIRD_EDIT
    // ... existing code ...

    You should still bias towards repeating as few lines of the original file as possible to convey the
    change. But, each edit should contain minimally sufficient context of unchanged lines around the
    code you're editing to resolve ambiguity.

    DO NOT omit spans of pre-existing code (or comments) without using the \"// ... existing code ...\"
    comment to indicate its absence. If you omit the existing code comment, the model may inadvertently
    delete these lines.

    If you plan on deleting a section, you must provide context before and after to delete it. If the
    initial code is \"Block 1\nBlock 2\nBlock 3\", and you want to remove Block 2, you would output \"//
    ... existing code ...\nBlock 1\nBlock 3\n// ... existing code ...\".

    Make sure it is clear what the edit should be, and where it should be applied. Make edits to a file
    in a single edit_file call instead of multiple edit_file calls to the same file. The apply model can
    handle many distinct edits at once.

    Args:
        path (str):
        body (ApplyEditRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[ApplyEditResponse, ErrorResponse]]
    """

    kwargs = _get_kwargs(
        path=path,
        body=body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    path: str,
    *,
    client: Union[Client],
    body: ApplyEditRequest,
) -> Union[ApplyEditResponse, ErrorResponse] | None:
    r"""Apply code edit

     Uses the configured LLM provider (Relace or Morph) to apply a code edit to the original content.

    To use this endpoint as an agent tool, follow these guidelines:

    Use this tool to make an edit to an existing file. This will be read by a less intelligent model,
    which will quickly apply the edit. You should make it clear what the edit is, while also minimizing
    the unchanged code you write.

    When writing the edit, you should specify each edit in sequence, with the special comment \"// ...
    existing code ...\" to represent unchanged code in between edited lines.

    Example format:
    // ... existing code ...
    FIRST_EDIT
    // ... existing code ...
    SECOND_EDIT
    // ... existing code ...
    THIRD_EDIT
    // ... existing code ...

    You should still bias towards repeating as few lines of the original file as possible to convey the
    change. But, each edit should contain minimally sufficient context of unchanged lines around the
    code you're editing to resolve ambiguity.

    DO NOT omit spans of pre-existing code (or comments) without using the \"// ... existing code ...\"
    comment to indicate its absence. If you omit the existing code comment, the model may inadvertently
    delete these lines.

    If you plan on deleting a section, you must provide context before and after to delete it. If the
    initial code is \"Block 1\nBlock 2\nBlock 3\", and you want to remove Block 2, you would output \"//
    ... existing code ...\nBlock 1\nBlock 3\n// ... existing code ...\".

    Make sure it is clear what the edit should be, and where it should be applied. Make edits to a file
    in a single edit_file call instead of multiple edit_file calls to the same file. The apply model can
    handle many distinct edits at once.

    Args:
        path (str):
        body (ApplyEditRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[ApplyEditResponse, ErrorResponse]
    """

    return sync_detailed(
        path=path,
        client=client,
        body=body,
    ).parsed


async def asyncio_detailed(
    path: str,
    *,
    client: Union[Client],
    body: ApplyEditRequest,
) -> Response[Union[ApplyEditResponse, ErrorResponse]]:
    r"""Apply code edit

     Uses the configured LLM provider (Relace or Morph) to apply a code edit to the original content.

    To use this endpoint as an agent tool, follow these guidelines:

    Use this tool to make an edit to an existing file. This will be read by a less intelligent model,
    which will quickly apply the edit. You should make it clear what the edit is, while also minimizing
    the unchanged code you write.

    When writing the edit, you should specify each edit in sequence, with the special comment \"// ...
    existing code ...\" to represent unchanged code in between edited lines.

    Example format:
    // ... existing code ...
    FIRST_EDIT
    // ... existing code ...
    SECOND_EDIT
    // ... existing code ...
    THIRD_EDIT
    // ... existing code ...

    You should still bias towards repeating as few lines of the original file as possible to convey the
    change. But, each edit should contain minimally sufficient context of unchanged lines around the
    code you're editing to resolve ambiguity.

    DO NOT omit spans of pre-existing code (or comments) without using the \"// ... existing code ...\"
    comment to indicate its absence. If you omit the existing code comment, the model may inadvertently
    delete these lines.

    If you plan on deleting a section, you must provide context before and after to delete it. If the
    initial code is \"Block 1\nBlock 2\nBlock 3\", and you want to remove Block 2, you would output \"//
    ... existing code ...\nBlock 1\nBlock 3\n// ... existing code ...\".

    Make sure it is clear what the edit should be, and where it should be applied. Make edits to a file
    in a single edit_file call instead of multiple edit_file calls to the same file. The apply model can
    handle many distinct edits at once.

    Args:
        path (str):
        body (ApplyEditRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[ApplyEditResponse, ErrorResponse]]
    """

    kwargs = _get_kwargs(
        path=path,
        body=body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    path: str,
    *,
    client: Union[Client],
    body: ApplyEditRequest,
) -> Union[ApplyEditResponse, ErrorResponse] | None:
    r"""Apply code edit

     Uses the configured LLM provider (Relace or Morph) to apply a code edit to the original content.

    To use this endpoint as an agent tool, follow these guidelines:

    Use this tool to make an edit to an existing file. This will be read by a less intelligent model,
    which will quickly apply the edit. You should make it clear what the edit is, while also minimizing
    the unchanged code you write.

    When writing the edit, you should specify each edit in sequence, with the special comment \"// ...
    existing code ...\" to represent unchanged code in between edited lines.

    Example format:
    // ... existing code ...
    FIRST_EDIT
    // ... existing code ...
    SECOND_EDIT
    // ... existing code ...
    THIRD_EDIT
    // ... existing code ...

    You should still bias towards repeating as few lines of the original file as possible to convey the
    change. But, each edit should contain minimally sufficient context of unchanged lines around the
    code you're editing to resolve ambiguity.

    DO NOT omit spans of pre-existing code (or comments) without using the \"// ... existing code ...\"
    comment to indicate its absence. If you omit the existing code comment, the model may inadvertently
    delete these lines.

    If you plan on deleting a section, you must provide context before and after to delete it. If the
    initial code is \"Block 1\nBlock 2\nBlock 3\", and you want to remove Block 2, you would output \"//
    ... existing code ...\nBlock 1\nBlock 3\n// ... existing code ...\".

    Make sure it is clear what the edit should be, and where it should be applied. Make edits to a file
    in a single edit_file call instead of multiple edit_file calls to the same file. The apply model can
    handle many distinct edits at once.

    Args:
        path (str):
        body (ApplyEditRequest):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[ApplyEditResponse, ErrorResponse]
    """

    return (
        await asyncio_detailed(
            path=path,
            client=client,
            body=body,
        )
    ).parsed




from .interpreter import CodeInterpreter
from .sandbox import (
    SandboxCodegen,
    SandboxFileSystem,
    SandboxInstance,
    SandboxPreviews,
    SandboxProcess,
)

__all__ = [
    "SandboxInstance",
    "SandboxFileSystem",
    "SandboxPreviews",
    "SandboxProcess",
    "SandboxCodegen",
    "CodeInterpreter",
]



import asyncio
import statistics
import time

from blaxel.core.sandbox.default import SandboxInstance
from blaxel.core.sandbox.default.action import SandboxAction

from utils import create_or_get_sandbox

SANDBOX_NAME = "fs-ls-benchmark"
NUM_ITERATIONS = 100


async def benchmark_fs_ls(sandbox: SandboxInstance, num_iterations: int = NUM_ITERATIONS):
    """Benchmark fs.ls performance with connection reuse."""
    print(f"\n{'='*60}")
    print(f"Benchmarking fs.ls with {num_iterations} iterations")
    print(f"{'='*60}\n")

    # Warm up - first request may be slower due to connection setup
    print("Warming up (first request)...")
    start = time.perf_counter()
    await sandbox.fs.ls("/")
    warmup_time = (time.perf_counter() - start) * 1000
    print(f"  Warmup request: {warmup_time:.2f}ms\n")

    # Run benchmark
    print(f"Running {num_iterations} fs.ls requests...")
    times_ms = []

    for i in range(num_iterations):
        start = time.perf_counter()
        await sandbox.fs.ls("/")
        elapsed_ms = (time.perf_counter() - start) * 1000
        times_ms.append(elapsed_ms)

        # Progress indicator every 10 iterations
        if (i + 1) % 10 == 0:
            print(f"  Completed {i + 1}/{num_iterations} requests...")

    # Calculate statistics
    total_time = sum(times_ms)
    avg_time = statistics.mean(times_ms)
    median_time = statistics.median(times_ms)
    min_time = min(times_ms)
    max_time = max(times_ms)
    std_dev = statistics.stdev(times_ms) if len(times_ms) > 1 else 0

    # Percentiles
    sorted_times = sorted(times_ms)
    p50 = sorted_times[int(len(sorted_times) * 0.50)]
    p90 = sorted_times[int(len(sorted_times) * 0.90)]
    p95 = sorted_times[int(len(sorted_times) * 0.95)]
    p99 = sorted_times[int(len(sorted_times) * 0.99)]

    print(f"\n{'='*60}")
    print("RESULTS")
    print(f"{'='*60}")
    print(f"Total requests:     {num_iterations}")
    print(f"Total time:         {total_time:.2f}ms ({total_time/1000:.2f}s)")
    print(f"Requests/second:    {num_iterations / (total_time / 1000):.2f}")
    print()
    print(f"Average:            {avg_time:.2f}ms")
    print(f"Median:             {median_time:.2f}ms")
    print(f"Min:                {min_time:.2f}ms")
    print(f"Max:                {max_time:.2f}ms")
    print(f"Std Dev:            {std_dev:.2f}ms")
    print()
    print("Percentiles:")
    print(f"  p50:              {p50:.2f}ms")
    print(f"  p90:              {p90:.2f}ms")
    print(f"  p95:              {p95:.2f}ms")
    print(f"  p99:              {p99:.2f}ms")
    print(f"{'='*60}\n")

    return {
        "total_requests": num_iterations,
        "total_time_ms": total_time,
        "requests_per_second": num_iterations / (total_time / 1000),
        "avg_ms": avg_time,
        "median_ms": median_time,
        "min_ms": min_time,
        "max_ms": max_time,
        "std_dev_ms": std_dev,
        "p50_ms": p50,
        "p90_ms": p90,
        "p95_ms": p95,
        "p99_ms": p99,
        "all_times_ms": times_ms,
    }


async def benchmark_concurrent_fs_ls(sandbox: SandboxInstance, num_concurrent: int = 10, num_batches: int = 10):
    """Benchmark concurrent fs.ls requests."""
    print(f"\n{'='*60}")
    print(f"Benchmarking CONCURRENT fs.ls")
    print(f"  {num_concurrent} concurrent requests x {num_batches} batches = {num_concurrent * num_batches} total")
    print(f"{'='*60}\n")

    all_times_ms = []

    for batch in range(num_batches):
        start = time.perf_counter()
        # Run concurrent requests
        await asyncio.gather(*[sandbox.fs.ls("/") for _ in range(num_concurrent)])
        batch_time_ms = (time.perf_counter() - start) * 1000
        all_times_ms.append(batch_time_ms)
        print(f"  Batch {batch + 1}/{num_batches}: {batch_time_ms:.2f}ms for {num_concurrent} concurrent requests")

    total_requests = num_concurrent * num_batches
    total_time = sum(all_times_ms)
    avg_batch_time = statistics.mean(all_times_ms)

    print(f"\n{'='*60}")
    print("CONCURRENT RESULTS")
    print(f"{'='*60}")
    print(f"Total requests:     {total_requests}")
    print(f"Total time:         {total_time:.2f}ms ({total_time/1000:.2f}s)")
    print(f"Requests/second:    {total_requests / (total_time / 1000):.2f}")
    print(f"Avg batch time:     {avg_batch_time:.2f}ms (for {num_concurrent} concurrent)")
    print(f"Avg per request:    {avg_batch_time / num_concurrent:.2f}ms (effective)")
    print(f"{'='*60}\n")

    return {
        "total_requests": total_requests,
        "total_time_ms": total_time,
        "requests_per_second": total_requests / (total_time / 1000),
        "avg_batch_time_ms": avg_batch_time,
        "effective_per_request_ms": avg_batch_time / num_concurrent,
    }


async def main():
    print("Starting fs.ls benchmark test")
    print("This will test HTTP client connection reuse performance\n")

    sandbox = await create_or_get_sandbox(SANDBOX_NAME)
    print(f"Using sandbox: {SANDBOX_NAME}")

    try:
        # Sequential benchmark
        sequential_results = await benchmark_fs_ls(sandbox, NUM_ITERATIONS)

        # Concurrent benchmark
        concurrent_results = await benchmark_concurrent_fs_ls(sandbox, num_concurrent=10, num_batches=10)

        # Summary
        print(f"\n{'='*60}")
        print("SUMMARY")
        print(f"{'='*60}")
        print(f"Sequential ({NUM_ITERATIONS} requests):")
        print(f"  - {sequential_results['requests_per_second']:.2f} req/s")
        print(f"  - {sequential_results['avg_ms']:.2f}ms avg latency")
        print()
        print(f"Concurrent (10 concurrent x 10 batches = 100 requests):")
        print(f"  - {concurrent_results['requests_per_second']:.2f} req/s")
        print(f"  - {concurrent_results['effective_per_request_ms']:.2f}ms effective latency")
        print(f"{'='*60}")

    finally:
        # Cleanup clients
        await SandboxAction.close_all_clients()
        print("\nClosed all HTTP clients")


if __name__ == "__main__":
    asyncio.run(main())



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="ResourceLogChart")


@_attrs_define
class ResourceLogChart:
    """Chart for a resource log

    Attributes:
        count (Union[Unset, int]): Count of the log
        debug (Union[Unset, int]): Debug count of the log
        error (Union[Unset, int]): Error count of the log
        fatal (Union[Unset, int]): Fatal count of the log
        info (Union[Unset, int]): Info count of the log
        timestamp (Union[Unset, str]): Timestamp of the log
        trace (Union[Unset, int]): Trace count of the log
        unknown (Union[Unset, int]): Unknown count of the log
        warning (Union[Unset, int]): Warning count of the log
    """

    count: Union[Unset, int] = UNSET
    debug: Union[Unset, int] = UNSET
    error: Union[Unset, int] = UNSET
    fatal: Union[Unset, int] = UNSET
    info: Union[Unset, int] = UNSET
    timestamp: Union[Unset, str] = UNSET
    trace: Union[Unset, int] = UNSET
    unknown: Union[Unset, int] = UNSET
    warning: Union[Unset, int] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        count = self.count

        debug = self.debug

        error = self.error

        fatal = self.fatal

        info = self.info

        timestamp = self.timestamp

        trace = self.trace

        unknown = self.unknown

        warning = self.warning

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if count is not UNSET:
            field_dict["count"] = count
        if debug is not UNSET:
            field_dict["debug"] = debug
        if error is not UNSET:
            field_dict["error"] = error
        if fatal is not UNSET:
            field_dict["fatal"] = fatal
        if info is not UNSET:
            field_dict["info"] = info
        if timestamp is not UNSET:
            field_dict["timestamp"] = timestamp
        if trace is not UNSET:
            field_dict["trace"] = trace
        if unknown is not UNSET:
            field_dict["unknown"] = unknown
        if warning is not UNSET:
            field_dict["warning"] = warning

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        count = d.pop("count", UNSET)

        debug = d.pop("debug", UNSET)

        error = d.pop("error", UNSET)

        fatal = d.pop("fatal", UNSET)

        info = d.pop("info", UNSET)

        timestamp = d.pop("timestamp", UNSET)

        trace = d.pop("trace", UNSET)

        unknown = d.pop("unknown", UNSET)

        warning = d.pop("warning", UNSET)

        resource_log_chart = cls(
            count=count,
            debug=debug,
            error=error,
            fatal=fatal,
            info=info,
            timestamp=timestamp,
            trace=trace,
            unknown=unknown,
            warning=warning,
        )

        resource_log_chart.additional_properties = d
        return resource_log_chart

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import asyncio
from logging import getLogger

from utils import create_or_get_sandbox

from blaxel.core.sandbox import SandboxInstance
from blaxel.core.sandbox.client.models.process_request import ProcessRequest

logger = getLogger(__name__)

SANDBOX_NAME = "sandbox-test-python-process-kill"


async def test_quick_process_kill(sandbox: SandboxInstance):
    """Test killing a quick process."""
    print("🔧 Testing quick process kill...")

    quick_process_name = "quick-test"
    quick_command = "echo 'Quick process'"

    # Start process
    process_request = ProcessRequest(name=quick_process_name, command=quick_command)
    await sandbox.process.exec(process_request)

    # Try to kill it (might already be completed)
    try:
        await sandbox.process.kill(quick_process_name)
        print("✅ Quick process kill successful")
    except Exception as e:
        print(f"⚠️ Quick process kill failed (expected if already completed): {e}")

    # Check final status
    try:
        final_process = await sandbox.process.get(quick_process_name)
        print(f"Final process status: {final_process.status}")
    except Exception as e:
        print(f"Could not get final process status: {e}")


async def test_stop_vs_kill(sandbox: SandboxInstance):
    """Test the difference between stop and kill."""
    print("🔧 Testing stop vs kill...")

    stop_process_name = "stop-test"
    stop_test_command = "sh -c 'trap \"echo Caught SIGTERM; exit 0\" TERM; echo Starting process; sleep 30; echo Process completed'"

    # Start long-running process
    process_request = ProcessRequest(name=stop_process_name, command=stop_test_command)
    await sandbox.process.exec(process_request)

    # Wait a bit for process to start
    await asyncio.sleep(2)

    # Test stop (graceful termination)
    print("Testing graceful stop...")
    try:
        await sandbox.process.stop(stop_process_name)
        print("✅ Process stop successful")

        # Wait for process to handle the signal
        await asyncio.sleep(2)

        # Check if process stopped gracefully
        stopped_process = await sandbox.process.get(stop_process_name)
        print(f"Stopped process status: {stopped_process.status}")
        print(f"Stopped process exit code: {stopped_process.exit_code}")

        # Get logs to see if it caught the signal
        logs = await sandbox.process.logs(stop_process_name)
        if "Caught SIGTERM" in logs:
            print("✅ Process handled SIGTERM gracefully")
        else:
            print("⚠️ Process may not have handled SIGTERM")

    except Exception as e:
        print(f"❌ Stop test failed: {e}")


async def test_force_kill(sandbox: SandboxInstance):
    """Test force killing a stubborn process."""
    print("🔧 Testing force kill...")

    kill_process_name = "kill-test"
    # Process that ignores SIGTERM
    kill_command = "sh -c 'trap \"\" TERM; echo Starting stubborn process; while true; do echo Still running...; sleep 2; done'"

    # Start stubborn process
    process_request = ProcessRequest(name=kill_process_name, command=kill_command)
    await sandbox.process.exec(process_request)

    # Wait for process to start
    await asyncio.sleep(3)

    # Check process is running
    running_process = await sandbox.process.get(kill_process_name)
    print(f"Process before kill: {running_process.status}")

    # Force kill the process
    try:
        await sandbox.process.kill(kill_process_name)
        print("✅ Force kill command sent")

        # Wait for kill to take effect
        await asyncio.sleep(2)

        # Check final status
        killed_process = await sandbox.process.get(kill_process_name)
        print(f"Process after kill: {killed_process.status}")
        print(f"Exit code: {killed_process.exit_code}")
    except Exception as e:
        print(f"❌ Force kill test failed: {e}")


async def main():
    """Main process kill test function."""
    print("🚀 Starting sandbox process kill tests...")

    try:
        # Create or get sandbox
        sandbox = await create_or_get_sandbox(SANDBOX_NAME)
        print(f"✅ Sandbox ready: {sandbox.metadata.name}")

        # Run different kill tests
        await test_quick_process_kill(sandbox)
        print()

        await test_stop_vs_kill(sandbox)
        print()

        await test_force_kill(sandbox)
        print()

        # Test process listing
        print("🔧 Testing process listing...")
        processes = await sandbox.process.list()
        print(f"Total processes: {len(processes)}")
        for proc in processes[-5:]:  # Show last 5 processes
            print(f"  - {proc.name}: {proc.status} (PID: {proc.pid})")

        print("🎉 All process kill tests completed successfully!")

    except Exception as e:
        print(f"❌ Process kill test failed with error: {e}")
        logger.exception("Process kill test error")
        raise
    finally:
        print("🧹 Final cleanup...")
        try:
            await SandboxInstance.delete(SANDBOX_NAME)
            print("✅ Sandbox deleted")
        except Exception as e:
            print(f"⚠️ Failed to delete sandbox: {e}")


if __name__ == "__main__":
    asyncio.run(main())
